#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.safe_sar_l2_ocn module."""
import unittest
import unittest.mock as mock

import numpy as np
import xarray as xr

from satpy.tests.utils import make_dataid


class TestSAFENC(unittest.TestCase):
    """Test various SAFE SAR L2 OCN file handlers."""

    @mock.patch("satpy.readers.safe_sar_l2_ocn.xr")
    def setUp(self, xr_):
        """Set up the tests."""
        from satpy.readers.safe_sar_l2_ocn import SAFENC

        self.channels = ["owiWindSpeed", "owiLon", "owiLat", "owiHs", "owiNrcs", "foo",
                         "owiPolarisationName", "owiCalConstObsi"]
        # Mock file access to return a fake dataset.
        self.dummy3d = np.zeros((2, 2, 1))
        self.dummy2d = np.zeros((2, 2))
        self.dummy1d = np.zeros((2))
        self.band = 1
        self.nc = xr.Dataset(
            {"owiWindSpeed": xr.DataArray(self.dummy2d, dims=("owiAzSize", "owiRaSize"), attrs={"_FillValue": np.nan}),
             "owiLon": xr.DataArray(data=self.dummy2d, dims=("owiAzSize", "owiRaSize")),
             "owiLat": xr.DataArray(data=self.dummy2d, dims=("owiAzSize", "owiRaSize")),
             "owiHs": xr.DataArray(data=self.dummy3d, dims=("owiAzSize", "owiRaSize", "oswPartition")),
             "owiNrcs": xr.DataArray(data=self.dummy3d, dims=("owiAzSize", "owiRaSize", "oswPolarization")),
             "foo": xr.DataArray(self.dummy2d, dims=("owiAzSize", "owiRaSize")),
             "owiPolarisationName": xr.DataArray(self.dummy1d, dims=("owiPolarisation")),
             "owiCalConstObsi": xr.DataArray(self.dummy1d, dims=("owiIncSize"))
             },
            attrs={"_FillValue": np.nan,
                   "missionName": "S1A"})
        xr_.open_dataset.return_value = self.nc

        # Instantiate reader using the mocked open_dataset() method. Also, make
        # the reader believe all abstract methods have been implemented.
        self.reader = SAFENC(filename="dummy",
                             filename_info={"start_time": 0,
                                            "end_time": 0,
                                            "fstart_time": 0,
                                            "fend_time": 0,
                                            "polarization": "vv"},
                             filetype_info={})

    def test_init(self):
        """Test reader initialization."""
        assert self.reader.start_time == 0
        assert self.reader.end_time == 0
        assert self.reader.fstart_time == 0
        assert self.reader.fend_time == 0

    def test_get_dataset(self):
        """Test getting a dataset."""
        for ch in self.channels:
            dt = self.reader.get_dataset(
                key=make_dataid(name=ch), info={})
            # ... this only compares the valid (unmasked) elements
            assert np.all(self.nc[ch] == dt.to_masked_array()), f"get_dataset() returns invalid data for dataset {ch}"



# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the writer base module."""
from __future__ import annotations

import datetime as dt
import os
import shutil

import pytest
import xarray as xr
from dask import array as da


class TestBaseWriter:
    """Test the base writer class."""

    def setup_method(self):
        """Set up tests."""
        import tempfile

        from pyresample.geometry import AreaDefinition

        from satpy.scene import Scene

        adef = AreaDefinition(
            "test", "test", "test", "EPSG:4326",
            100, 200, (-180., -90., 180., 90.),
        )
        ds1 = xr.DataArray(
            da.arange(100 * 200).reshape((100, 200)).rechunk(50),
            dims=("y", "x"),
            attrs={
                "name": "test",
                "start_time": dt.datetime(2018, 1, 1, 0, 0, 0),
                "sensor": "fake_sensor",
                "area": adef,
            }
        )
        ds2 = ds1.copy()
        ds2.attrs["sensor"] = {"fake_sensor1", "fake_sensor2"}
        self.scn = Scene()
        self.scn["test"] = ds1
        self.scn["test2"] = ds2

        # Temp dir
        self.base_dir = tempfile.mkdtemp()

    def teardown_method(self):
        """Remove the temporary directory created for a test."""
        try:
            shutil.rmtree(self.base_dir, ignore_errors=True)
        except OSError:
            pass

    def test_save_dataset_static_filename(self):
        """Test saving a dataset with a static filename specified."""
        self.scn.save_datasets(datasets=["test"], base_dir=self.base_dir, filename="geotiff.tif")
        assert os.path.isfile(os.path.join(self.base_dir, "geotiff.tif"))

    @pytest.mark.parametrize(
        ("fmt_fn", "exp_fns"),
        [
            ("geotiff_{name}_{start_time:%Y%m%d_%H%M%S}.tif",
             ["geotiff_test_20180101_000000.tif", "geotiff_test2_20180101_000000.tif"]),
            ("geotiff_{name}_{sensor}.tif",
             ["geotiff_test_fake_sensor.tif", "geotiff_test2_fake_sensor1-fake_sensor2.tif"]),
        ]
    )
    def test_save_dataset_dynamic_filename(self, fmt_fn, exp_fns):
        """Test saving a dataset with a format filename specified."""
        self.scn.save_datasets(base_dir=self.base_dir, filename=fmt_fn)
        for exp_fn in exp_fns:
            exp_path = os.path.join(self.base_dir, exp_fn)
            assert os.path.isfile(exp_path)

    def test_save_dataset_dynamic_filename_with_dir(self):
        """Test saving a dataset with a format filename that includes a directory."""
        fmt_fn = os.path.join("{start_time:%Y%m%d}", "geotiff_{name}_{start_time:%Y%m%d_%H%M%S}.tif")
        exp_fn = os.path.join("20180101", "geotiff_test_20180101_000000.tif")
        self.scn.save_datasets(base_dir=self.base_dir, filename=fmt_fn)
        assert os.path.isfile(os.path.join(self.base_dir, exp_fn))

        # change the filename pattern but keep the same directory
        fmt_fn2 = os.path.join("{start_time:%Y%m%d}", "geotiff_{name}_{start_time:%Y%m%d_%H}.tif")
        exp_fn2 = os.path.join("20180101", "geotiff_test_20180101_00.tif")
        self.scn.save_datasets(base_dir=self.base_dir, filename=fmt_fn2)
        assert os.path.isfile(os.path.join(self.base_dir, exp_fn2))
        # the original file should still exist
        assert os.path.isfile(os.path.join(self.base_dir, exp_fn))



"""Unit tests for GMS-5 VISSR navigation."""

import numpy as np
import pytest

from satpy.tests.reader_tests.utils import get_jit_methods
from satpy.tests.utils import skip_numba_unstable_if_missing

try:
    import satpy.readers.gms.gms5_vissr_navigation as nav
except ImportError as err:
    if skip_numba_unstable_if_missing():
        pytest.skip(f"Numba is not compatible with unstable NumPy: {err!s}", allow_module_level=True)
    raise

# Navigation references computed with JMA's Msial library (files
# VISSR_19960217_2331_IR1.A.IMG and VISSR_19960217_2331_VIS.A.IMG). The VIS
# navigation is slightly off (< 0.01 deg) compared to JMA's reference.
# This is probably due to precision problems with the copied numbers.
IR_NAVIGATION_REFERENCE = [
    {
        "pixel": nav.Pixel(line=686, pixel=1680),
        "lon": 139.990380,
        "lat": 35.047056,
        "nav_params": nav.PixelNavigationParameters(
            attitude=nav.Attitude(
                angle_between_earth_and_sun=3.997397917902958,
                angle_between_sat_spin_and_z_axis=3.149118633034304,
                angle_between_sat_spin_and_yz_plane=0.000546042025980,
            ),
            orbit=nav.Orbit(
                angles=nav.OrbitAngles(
                    greenwich_sidereal_time=2.468529732418296,
                    declination_from_sat_to_sun=-0.208770861178982,
                    right_ascension_from_sat_to_sun=3.304369303579407,
                ),
                sat_position=nav.Vector3D(
                    x=-32390963.148471601307392,
                    y=27003395.381247851997614,
                    z=-228134.860026293463307,
                ),
                nutation_precession=np.array(
                    [[0.999936381496146, -0.010344758016410, -0.004496547784299],
                     [0.010344942303489, 0.999946489495557, 0.000017727054455],
                     [0.004496123789670, -0.000064242454080, 0.999989890320785]]
                ),
            ),
            proj_params=nav.ProjectionParameters(
                image_offset=nav.ImageOffset(
                    line_offset=1378.5,
                    pixel_offset=1672.5,
                ),
                scanning_angles=nav.ScanningAngles(
                    stepping_angle=0.000140000047395,
                    sampling_angle=0.000095719995443,
                    misalignment=np.array(
                        [[0.999999165534973, 0.000510364072397, 0.001214201096445],
                         [-0.000511951977387, 0.999999046325684, 0.001307720085606],
                         [-0.001213532872498, -0.001308340579271, 0.999998450279236]]
                    )
                ),
                earth_ellipsoid=nav.EarthEllipsoid(
                    flattening=0.003352813177897,
                    equatorial_radius=6378136.0
                )
            ),
        )
    },
    {
        "pixel": nav.Pixel(line=2089, pixel=1793),
        "lon": 144.996967,
        "lat": -34.959853,
        "nav_params": nav.PixelNavigationParameters(
            attitude=nav.Attitude(
                angle_between_earth_and_sun=3.935707944355762,
                angle_between_sat_spin_and_z_axis=3.149118633034304,
                angle_between_sat_spin_and_yz_plane=0.000546042025980,
            ),
            orbit=nav.Orbit(
                angles=nav.OrbitAngles(
                    greenwich_sidereal_time=2.530392320846865,
                    declination_from_sat_to_sun=-0.208713576872247,
                    right_ascension_from_sat_to_sun=3.242660398458377,
                ),
                sat_position=nav.Vector3D(
                    x=-32390273.633551981300116,
                    y=27003859.543135114014149,
                    z=-210800.087589388160268,
                ),
                nutation_precession=np.array(
                    [[0.999936381432029, -0.010344763228876, -0.004496550050695],
                     [0.010344947502662, 0.999946489441823, 0.000017724053657],
                     [0.004496126086653, -0.000064239500295, 0.999989890310647]]
                ),
            ),
            proj_params=nav.ProjectionParameters(
                image_offset=nav.ImageOffset(
                    line_offset=1378.5,
                    pixel_offset=1672.5,
                ),
                scanning_angles=nav.ScanningAngles(
                    stepping_angle=0.000140000047395,
                    sampling_angle=0.000095719995443,
                    misalignment=np.array(
                        [[0.999999165534973, 0.000510364072397, 0.001214201096445],
                         [-0.000511951977387, 0.999999046325684, 0.001307720085606],
                         [-0.001213532872498, -0.001308340579271, 0.999998450279236]]
                    ),
                ),
                earth_ellipsoid=nav.EarthEllipsoid(
                    flattening=0.003352813177897,
                    equatorial_radius=6378136
                )
            ),
        )
    }
]


VIS_NAVIGATION_REFERENCE = [
    {
        "pixel": nav.Pixel(line=2744, pixel=6720),
        "lon": 139.975527,
        "lat": 35.078028,
        "nav_params": nav.PixelNavigationParameters(
            attitude=nav.Attitude(
                angle_between_earth_and_sun=3.997397918405798,
                angle_between_sat_spin_and_z_axis=3.149118633034304,
                angle_between_sat_spin_and_yz_plane=0.000546042025980,
            ),
            orbit=nav.Orbit(
                angles=nav.OrbitAngles(
                    greenwich_sidereal_time=2.468529731914041,
                    declination_from_sat_to_sun=-0.208770861179448,
                    right_ascension_from_sat_to_sun=3.304369304082406,
                ),
                sat_position=nav.Vector3D(
                    x=-32390963.148477241396904,
                    y=27003395.381243918091059,
                    z=-228134.860164520738181,
                ),
                nutation_precession=np.array(
                    [[0.999936381496146, -0.010344758016410, -0.004496547784299],
                     [0.010344942303489, 0.999946489495557, 0.000017727054455],
                     [0.004496123789670, -0.000064242454080, 0.999989890320785]]
                ),
            ),
            proj_params=nav.ProjectionParameters(
                image_offset=nav.ImageOffset(
                    line_offset=5513.0,
                    pixel_offset=6688.5,
                ),
                scanning_angles=nav.ScanningAngles(
                    stepping_angle=0.000035000004573,
                    sampling_angle=0.000023929998861,
                    misalignment=np.array(
                        [[0.999999165534973, 0.000510364072397, 0.001214201096445],
                         [-0.000511951977387, 0.999999046325684, 0.001307720085606],
                         [-0.001213532872498, -0.001308340579271, 0.999998450279236]]
                    ),
                ),
                earth_ellipsoid=nav.EarthEllipsoid(
                    flattening=0.003352813177897,
                    equatorial_radius=6378136
                )
            ),
        )
    },
    {
        "pixel": nav.Pixel(line=8356, pixel=7172),
        "lon": 144.980104,
        "lat": -34.929123,
        "nav_params": nav.PixelNavigationParameters(
            attitude=nav.Attitude(
                angle_between_earth_and_sun=3.935707944858620,
                angle_between_sat_spin_and_z_axis=3.149118633034304,
                angle_between_sat_spin_and_yz_plane=0.000546042025980,
            ),
            orbit=nav.Orbit(
                angles=nav.OrbitAngles(
                    greenwich_sidereal_time=2.530392320342610,
                    declination_from_sat_to_sun=-0.208713576872715,
                    right_ascension_from_sat_to_sun=3.242660398961383,
                ),
                sat_position=nav.Vector3D(
                    x=-32390273.633557569235563,
                    y=27003859.543131537735462,
                    z=-210800.087734811415430,
                ),
                nutation_precession=np.array(
                    [[0.999936381432029, -0.010344763228876, -0.004496550050695],
                     [0.010344947502662, 0.999946489441823, 0.000017724053657],
                     [0.004496126086653, -0.000064239500295, 0.999989890310647]]
                ),
            ),
            proj_params=nav.ProjectionParameters(
                image_offset=nav.ImageOffset(
                    line_offset=5513.0,
                    pixel_offset=6688.5,
                ),
                scanning_angles=nav.ScanningAngles(
                    stepping_angle=0.000035000004573,
                    sampling_angle=0.000023929998861,
                    misalignment=np.array(
                        [[0.999999165534973, 0.000510364072397, 0.001214201096445],
                         [-0.000511951977387, 0.999999046325684, 0.001307720085606],
                         [-0.001213532872498, -0.001308340579271, 0.999998450279236]]
                    ),
                ),
                earth_ellipsoid=nav.EarthEllipsoid(
                    flattening=0.003352813177897,
                    equatorial_radius=6378136
                )
            ),
        )
    },
]

NAVIGATION_REFERENCE = VIS_NAVIGATION_REFERENCE + IR_NAVIGATION_REFERENCE


@pytest.fixture(params=[False, True], autouse=True)
def _disable_jit(request, monkeypatch):
    """Run tests with jit enabled and disabled.

    Reason: Coverage report is only accurate with jit disabled.
    """
    if request.param:
        jit_methods = get_jit_methods(nav)
        for name, method in jit_methods.items():
            monkeypatch.setattr(name, method.py_func)


class TestSinglePixelNavigation:
    """Test navigation of a single pixel."""

    @pytest.mark.parametrize(
        ("point", "nav_params", "expected"),
        [
            (ref["pixel"], ref["nav_params"], (ref["lon"], ref["lat"]))
            for ref in NAVIGATION_REFERENCE
        ],
    )
    def test_get_lon_lat(self, point, nav_params, expected):
        """Test getting lon/lat coordinates for a given pixel."""
        lon, lat = nav.get_lon_lat(point, nav_params)
        np.testing.assert_allclose((lon, lat), expected)

    def test_transform_image_coords_to_scanning_angles(self):
        """Test transformation from image coordinates to scanning angles."""
        offset = nav.ImageOffset(line_offset=100, pixel_offset=200)
        scanning_angles = nav.ScanningAngles(
            stepping_angle=0.01, sampling_angle=0.02, misalignment=-999
        )
        angles = nav.transform_image_coords_to_scanning_angles(
            point=nav.Pixel(199, 99),
            image_offset=offset,
            scanning_angles=scanning_angles,
        )
        np.testing.assert_allclose(angles, [-2, 1])

    def test_transform_scanning_angles_to_satellite_coords(self):
        """Test transformation from scanning angles to satellite coordinates."""
        scanning_angles = nav.Vector2D(np.pi, np.pi / 2)
        misalignment = np.diag([1, 2, 3]).astype(float)
        point_sat = nav.transform_scanning_angles_to_satellite_coords(
            scanning_angles, misalignment
        )
        np.testing.assert_allclose(point_sat, [0, 0, 3], atol=1e-12)

    def test_transform_satellite_to_earth_fixed_coords(self):
        """Test transformation from satellite to earth-fixed coordinates."""
        point_sat = nav.Vector3D(1, 2, 3)
        attitude = nav.Attitude(
            angle_between_earth_and_sun=np.pi,
            angle_between_sat_spin_and_z_axis=np.pi,
            angle_between_sat_spin_and_yz_plane=np.pi / 2,
        )
        orbit = nav.Orbit(
            angles=nav.OrbitAngles(
                greenwich_sidereal_time=np.pi,
                declination_from_sat_to_sun=np.pi,
                right_ascension_from_sat_to_sun=np.pi / 2,
            ),
            sat_position=nav.Vector3D(-999, -999, -999),
            nutation_precession=np.diag([1, 2, 3]).astype(float),
        )
        res = nav.transform_satellite_to_earth_fixed_coords(point_sat, orbit, attitude)
        np.testing.assert_allclose(res, [-3, 1, -2])

    def test_intersect_view_vector_with_earth(self):
        """Test intersection of a view vector with the earth's surface."""
        view_vector = nav.Vector3D(-1, 0, 0)
        ellipsoid = nav.EarthEllipsoid(equatorial_radius=6371 * 1000, flattening=0.003)
        sat_pos = nav.Vector3D(x=36000 * 1000.0, y=0.0, z=0.0)
        point = nav.intersect_with_earth(view_vector, sat_pos, ellipsoid)
        exp = [ellipsoid.equatorial_radius, 0, 0]
        np.testing.assert_allclose(point, exp)

    @pytest.mark.parametrize(
        ("point_earth_fixed", "point_geodetic_exp"),
        [
            ([0, 0, 1], [0, 90]),
            ([0, 0, -1], [0, -90]),
            ([1, 0, 0], [0, 0]),
            ([-1, 0, 0], [180, 0]),
            ([1, 1, 1], [45, 35.426852]),
        ],
    )
    def test_transform_earth_fixed_to_geodetic_coords(
        self, point_earth_fixed, point_geodetic_exp
    ):
        """Test transformation from earth-fixed to geodetic coordinates."""
        point_geodetic = nav.transform_earth_fixed_to_geodetic_coords(
            nav.Vector3D(*point_earth_fixed),
            0.003
        )
        np.testing.assert_allclose(point_geodetic, point_geodetic_exp)

    def test_normalize_vector(self):
        """Test vector normalization."""
        v = nav.Vector3D(1, 2, 3)
        norm = np.sqrt(14)
        exp = nav.Vector3D(1 / norm, 2 / norm, 3 / norm)
        normed = nav.normalize_vector(v)
        np.testing.assert_allclose(normed, exp)


class TestImageNavigation:
    """Test navigation of an entire image."""

    @pytest.fixture
    def expected(self):
        """Get expected coordinates."""
        exp = {
            "lon": [[-114.56923, -112.096837, -109.559702],
                    [8.33221, 8.793893, 9.22339],
                    [15.918476, 16.268354, 16.6332]],
            "lat": [[-23.078721, -24.629845, -26.133314],
                    [-42.513409, -39.790231, -37.06392],
                    [3.342834, 6.07043, 8.795932]]
        }
        return exp

    def test_get_lons_lats(self, navigation_params, expected):
        """Test getting lon/lat coordinates."""
        lons, lats = nav.get_lons_lats(
            lines=np.array([1000, 1500, 2000]),
            pixels=np.array([1000, 1500, 2000]),
            nav_params=navigation_params,
        )
        np.testing.assert_allclose(lons, expected["lon"])
        np.testing.assert_allclose(lats, expected["lat"])


class TestPredictionInterpolation:
    """Test interpolation of orbit and attitude predictions."""

    @pytest.mark.parametrize(
        ("obs_time", "expected"), [(-1, np.nan), (1.5, 2.5), (5, np.nan)]
    )
    def test_interpolate_continuous(self, obs_time, expected):
        """Test interpolation of continuous variables."""
        prediction_times = np.array([0, 1, 2, 3])
        predicted_values = np.array([1, 2, 3, 4])
        res = nav.interpolate_continuous(obs_time, prediction_times, predicted_values)
        np.testing.assert_allclose(res, expected)

    @pytest.mark.parametrize(
        ("obs_time", "expected"),
        [
            (-1, np.nan),
            (1.5, 0.75 * np.pi),
            (2.5, -0.75 * np.pi),
            (3.5, -0.25 * np.pi),
            (5, np.nan),
        ],
    )
    def test_interpolate_angles(self, obs_time, expected):
        """Test interpolation of periodic angles."""
        prediction_times = np.array([0, 1, 2, 3, 4])
        predicted_angles = np.array(
            [0, 0.5 * np.pi, np.pi, 1.5 * np.pi, 2 * np.pi]
        )  # already unwrapped
        res = nav.interpolate_angles(obs_time, prediction_times, predicted_angles)
        np.testing.assert_allclose(res, expected)

    @pytest.mark.parametrize(
        ("obs_time", "expected"),
        [
            (-1, np.nan * np.ones((2, 2))),
            (1.5, [[1, 0], [0, 2]]),
            (3, np.nan * np.ones((2, 2))),
        ],
    )
    def test_interpolate_nearest(self, obs_time, expected):
        """Test nearest neighbour interpolation."""
        prediction_times = np.array([0, 1, 2])
        predicted_angles = np.array(
            [np.zeros((2, 2)), np.diag((1, 2)), np.zeros((2, 2))]
        )
        res = nav.interpolate_nearest(obs_time, prediction_times, predicted_angles)
        np.testing.assert_allclose(res, expected)

    def test_interpolate_orbit_prediction(
        self, obs_time, orbit_prediction, orbit_expected
    ):
        """Test interpolating orbit prediction."""
        orbit_prediction = orbit_prediction.to_numba()
        orbit = nav.interpolate_orbit_prediction(orbit_prediction, obs_time)
        _assert_namedtuple_close(orbit, orbit_expected)

    def test_interpolate_attitude_prediction(
        self, obs_time, attitude_prediction, attitude_expected
    ):
        """Test interpolating attitude prediction."""
        attitude_prediction = attitude_prediction.to_numba()
        attitude = nav.interpolate_attitude_prediction(attitude_prediction, obs_time)
        _assert_namedtuple_close(attitude, attitude_expected)

    @pytest.fixture
    def obs_time(self):
        """Get observation time."""
        return 2.5

    @pytest.fixture
    def orbit_expected(self):
        """Get expected orbit."""
        return nav.Orbit(
            angles=nav.OrbitAngles(
                greenwich_sidereal_time=1.5,
                declination_from_sat_to_sun=1.6,
                right_ascension_from_sat_to_sun=1.7,
            ),
            sat_position=nav.Vector3D(
                x=1.8,
                y=1.9,
                z=2.0,
            ),
            nutation_precession=1.6 * np.identity(3),
        )

    @pytest.fixture
    def attitude_expected(self):
        """Get expected attitude."""
        return nav.Attitude(
            angle_between_earth_and_sun=1.5,
            angle_between_sat_spin_and_z_axis=1.6,
            angle_between_sat_spin_and_yz_plane=1.7,
        )


@pytest.fixture
def sampling_angle():
    """Get sampling angle."""
    return 0.000095719995443


@pytest.fixture
def scan_params(sampling_angle):
    """Get scanning parameters."""
    return nav.ScanningParameters(
        start_time_of_scan=0,
        spinning_rate=0.5,
        num_sensors=1,
        sampling_angle=sampling_angle,
    )


@pytest.fixture
def attitude_prediction():
    """Get attitude prediction."""
    return nav.AttitudePrediction(
        prediction_times=np.array([1.0, 2.0, 3.0]),
        attitude=nav.Attitude(
            angle_between_earth_and_sun=np.array([0.0, 1.0, 2.0]),
            angle_between_sat_spin_and_z_axis=np.array([0.1, 1.1, 2.1]),
            angle_between_sat_spin_and_yz_plane=np.array([0.2, 1.2, 2.2]),
        ),
    )


@pytest.fixture
def orbit_prediction():
    """Get orbit prediction."""
    return nav.OrbitPrediction(
        prediction_times=np.array([1.0, 2.0, 3.0, 4.0]),
        angles=nav.OrbitAngles(
            greenwich_sidereal_time=np.array([0.0, 1.0, 2.0, 3.0]),
            declination_from_sat_to_sun=np.array([0.1, 1.1, 2.1, 3.1]),
            right_ascension_from_sat_to_sun=np.array([0.2, 1.2, 2.2, 3.2]),
        ),
        sat_position=nav.Vector3D(
            x=np.array([0.3, 1.3, 2.3, 3.3]),
            y=np.array([0.4, 1.4, 2.4, 3.4]),
            z=np.array([0.5, 1.5, 2.5, 3.5]),
        ),
        nutation_precession=np.array(
            [
                0.6 * np.identity(3),
                1.6 * np.identity(3),
                2.6 * np.identity(3),
                3.6 * np.identity(3),
            ]
        ),
    )


@pytest.fixture
def proj_params(sampling_angle):
    """Get projection parameters."""
    return nav.ProjectionParameters(
        image_offset=nav.ImageOffset(
            line_offset=1378.5,
            pixel_offset=1672.5,
        ),
        scanning_angles=nav.ScanningAngles(
            stepping_angle=0.000140000047395,
            sampling_angle=sampling_angle,
            misalignment=np.identity(3).astype(np.float64),
        ),
        earth_ellipsoid=nav.EarthEllipsoid(
            flattening=0.003352813177897,
            equatorial_radius=6378136,
        ),
    )


@pytest.fixture
def static_nav_params(proj_params, scan_params):
    """Get static navigation parameters."""
    return nav.StaticNavigationParameters(proj_params, scan_params)


@pytest.fixture
def predicted_nav_params(attitude_prediction, orbit_prediction):
    """Get predicted navigation parameters."""
    return nav.PredictedNavigationParameters(attitude_prediction, orbit_prediction)


@pytest.fixture
def navigation_params(static_nav_params, predicted_nav_params):
    """Get image navigation parameters."""
    return nav.ImageNavigationParameters(static_nav_params, predicted_nav_params)


def test_get_observation_time():
    """Test getting a pixel's observation time."""
    scan_params = nav.ScanningParameters(
        start_time_of_scan=50000.0,
        spinning_rate=100,
        num_sensors=1,
        sampling_angle=0.01,
    )
    pixel = nav.Pixel(11, 100)
    obs_time = nav.get_observation_time(pixel, scan_params)
    np.testing.assert_allclose(obs_time, 50000.0000705496871047)


def _assert_namedtuple_close(a, b):
    cls_name = b.__class__.__name__
    assert a.__class__ == b.__class__
    for attr in b._fields:
        a_attr = getattr(a, attr)
        b_attr = getattr(b, attr)
        if _is_namedtuple(b_attr):
            _assert_namedtuple_close(a_attr, b_attr)
        np.testing.assert_allclose(
            a_attr, b_attr, err_msg=f"{cls_name} attribute {attr} differs"
        )


def _is_namedtuple(obj):
    return hasattr(obj, "_fields")



#!/usr/bin/env python
# Copyright (c) 2016-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for reader base and utility modules that were moved to core sub-package raise warnings."""

import pytest


def test_abi_base_warns():
    """Test that there's a warning when importing from ABI base from the old location."""
    from satpy.readers import abi_base

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(abi_base, "NC_ABI_BASE")


def test_fci_base_warns():
    """Test that there's a warning when importing from FCI base from the old location."""
    from satpy.readers import fci_base

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(fci_base, "calculate_area_extent")


@pytest.mark.parametrize("name",
                         ["timecds2datetime",
                          "recarray2dict",
                          "get_service_mode",
                         ]
                         )
def test_eum_base_warns(name):
    """Test that there's a warning when importing from EUM base from the old location."""
    from satpy.readers import eum_base

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(eum_base, name)


def test_fy4_base_warns():
    """Test that there's a warning when importing from FY4 base from the old location."""
    from satpy.readers import fy4_base

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(fy4_base, "FY4Base")


@pytest.mark.parametrize("name",
                         ["from_sds",
                          "HDF4FileHandler",
                          "SDS",
                         ]
                         )
def test_hdf4_utils_warns(name):
    """Test that there's a warning when importing from hdf4 utils from the old location."""
    from satpy.readers import hdf4_utils

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(hdf4_utils, name)


@pytest.mark.parametrize("name",
                         ["from_h5_array",
                          "HDF5FileHandler",
                         ]
                         )
def test_hdf5_utils_warns(name):
    """Test that there's a warning when importing from hdf5 utils from the old location."""
    from satpy.readers import hdf5_utils

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(hdf5_utils, name)


@pytest.mark.parametrize("name",
                         ["interpolate",
                          "HDFEOSBaseFileReader",
                          "HDFEOSGeoReader",
                         ]
                         )
def test_hdfeos_base_warns(name):
    """Test that there's a warning when importing from hdfeos base from the old location."""
    from satpy.readers import hdfeos_base

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(hdfeos_base, name)


@pytest.mark.parametrize("name",
                         ["decompress_file",
                          "decompress_buffer",
                          "get_header_id",
                          "get_header_content",
                          "HRITFileHandler",
                          "HRITSegment",
                         ]
                         )
def test_hrit_base_warns(name):
    """Test that there's a warning when importing from HRIT base from the old location."""
    from satpy.readers import hrit_base

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(hrit_base, name)


@pytest.mark.parametrize("name",
                         ["LINCFileHandler",
                         ]
                         )
def test_li_base_nc_warns(name):
    """Test that there's a warning when importing from LI netCDF4 base from the old location."""
    from satpy.readers import li_base_nc

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(li_base_nc, name)


@pytest.mark.parametrize("name",
                         ["get_data_as_xarray",
                          "NetCDF4FileHandler",
                          "NetCDF4FsspecFileHandler",
                         ]
                         )
def test_netcdf_utils_warns(name):
    """Test that there's a warning when importing from netCDF4 utils from the old location."""
    from satpy.readers import netcdf_utils

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(netcdf_utils, name)


@pytest.mark.parametrize("name",
                         ["get_cds_time",
                          "add_scanline_acq_time",
                          "dec10216",
                          "chebyshev",
                          "chebyshev_3d",
                          "get_satpos",
                          "calculate_area_extent",
                          "create_coef_dict",
                          "get_padding_area",
                          "pad_data_horizontally",
                          "pad_data_vertically",
                          "mask_bad_quality",
                          "round_nom_time",
                          "MpefProductHeader",
                          "SEVIRICalibrationAlgorithm",
                          "SEVIRICalibrationHandler",
                          "NoValidOrbitParams",
                          "OrbitPolynomial",
                          "OrbitPolynomialFinder",
                          "NominalCoefficients",
                          "GsicsCoefficients",
                          "MeirinkCoefficients",
                         ]
                         )
def test_seviri_base_warns(name):
    """Test that there's a warning when importing from SEVIRI base from the old location."""
    from satpy.readers import seviri_base

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(seviri_base, name)


@pytest.mark.parametrize("name",
                         ["ViiNCBaseFileHandler",
                         ]
                         )
def test_vii_base_nc_warns(name):
    """Test that there's a warning when importing from VII NetCDF4 base from the old location."""
    from satpy.readers import vii_base_nc

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(vii_base_nc, name)


@pytest.mark.parametrize("name",
                         ["JPSS_SDR_FileHandler",
                          "DATASET_KEYS",
                          "ATMS_DATASET_KEYS",
                          "VIIRS_DATASET_KEYS",
                         ]
                         )
def test_viirs_atms_sdr_warns(name):
    """Test that there's a warning when importing from VIIRS ATMS SRD base from the old location."""
    from satpy.readers import viirs_atms_sdr_base

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(viirs_atms_sdr_base, name)


@pytest.mark.parametrize("name",
                         ["C1",
                          "C2",
                          "TIE_POINTS_FACTOR",
                          "SCAN_ALT_TIE_POINTS",
                          "MEAN_EARTH_RADIUS",
                         ]
                         )
def test_vii_utils_warns(name):
    """Test that there's a warning when importing from VII utils from the old location."""
    from satpy.readers import vii_utils

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(vii_utils, name)


@pytest.mark.parametrize("name",
                         ["listify_string",
                          "load_yaml_configs",
                          "split_integer_in_most_equal_parts",
                          "AbstractYAMLReader",
                          "GenericYAMLReader",
                          "FileYAMLReader",
                          "GEOFlippableFileYAMLReader",
                          "GEOSegmentYAMLReader",
                          "GEOVariableSegmentYAMLReader",
                         ]
                         )
def test_yaml_reader_warns(name):
    """Test that there's a warning when importing from YAML reader from the old location."""
    from satpy.readers import yaml_reader

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(yaml_reader, name)


@pytest.mark.parametrize("name",
                         ["process_delimiter",
                          "process_field",
                          "process_array",
                          "to_dtype",
                          "to_scaled_dtype",
                          "to_scales",
                          "parse_format",
                          "XMLFormat",
                         ]
                         )
def test_xmlformat_warns(name):
    """Test that there's a warning when importing from xmlformat from the old location."""
    from satpy.readers import xmlformat

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(xmlformat, name)


@pytest.mark.parametrize("name",
                         ["open_dataset",
                          "BaseFileHandler",
                         ]
                         )
def test_file_handlers_warns(name):
    """Test that there's a warning when importing from file_handlers from the old location."""
    from satpy.readers import file_handlers

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(file_handlers, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Shared utilities for correcting reflectance data using the 'crefl' algorithm.

The CREFL algorithm in this module is based on the `NASA CREFL SPA`_ software,
the `NASA CVIIRS SPA`_, and customizations of these algorithms for ABI/AHI by
Ralph Kuehn and Min Oo at the Space Science and Engineering Center (SSEC).

The CREFL SPA documentation page describes the algorithm by saying:

    The CREFL_SPA processes MODIS Aqua and Terra Level 1B DB data to create the
    MODIS Level 2 Corrected Reflectance product. The algorithm performs a simple
    atmospheric correction with MODIS visible, near-infrared, and short-wave
    infrared bands (bands 1 through 16).

    It corrects for molecular (Rayleigh) scattering and gaseous absorption (water
    vapor and ozone) using climatological values for gas contents. It requires no
    real-time input of ancillary data. The algorithm performs no aerosol
    correction. The Corrected Reflectance products created by CREFL_SPA are very
    similar to the MODIS Land Surface Reflectance product (MOD09) in clear
    atmospheric conditions, since the algorithms used to derive both are based on
    the 6S Radiative Transfer Model. The products show differences in the presence
    of aerosols, however, because the MODIS Land Surface Reflectance product uses
    a more complex atmospheric correction algorithm that includes a correction for
    aerosols.

The additional logic to support ABI (AHI support not included) was originally
written by Ralph Kuehn and Min Oo at SSEC. Additional modifications were
performed by Martin Raspaud, David Hoese, and Will Roberts to make the code
work together and be more dask compatible.

The AHI/ABI implementation is based on the MODIS collection 6 algorithm, where
a spherical-shell atmosphere was assumed rather than a plane-parallel. See
Appendix A in: "The Collection 6 MODIS aerosol products over land and ocean"
Atmos. Meas. Tech., 6, 2989–3034, 2013 www.atmos-meas-tech.net/6/2989/2013/
:doi:`10.5194/amt-6-2989-2013`.


The original CREFL code is similar to what is described in appendix A1 (page
74) of the ATBD for the `MODIS MOD04/MYD04`_ data product.

.. _NASA CREFL SPA: https://directreadout.sci.gsfc.nasa.gov/?id=dspContent&cid=92&type=software
.. _NASA CVIIRS SPA: https://directreadout.sci.gsfc.nasa.gov/?id=dspContent&cid=277&type=software
.. _MODIS MOD04/MYD04: https://modis.gsfc.nasa.gov/data/atbd/atbd_mod02.pdf


"""
from __future__ import annotations

import logging
from typing import Optional, Type, Union

import dask.array as da
import numpy as np
import xarray as xr

from satpy.dataset.dataid import WavelengthRange

LOG = logging.getLogger(__name__)

UO3_MODIS = 0.319
UH2O_MODIS = 2.93
UO3_VIIRS = 0.285
UH2O_VIIRS = 2.93

MAXSOLZ = 86.5
MAXAIRMASS = 18
SCALEHEIGHT = 8000
FILL_INT16 = 32767
TAUSTEP4SPHALB_ABI = .0003
TAUSTEP4SPHALB = .0001

MAXNUMSPHALBVALUES = 4000  # with no aerosol taur <= 0.4 in all bands everywhere
REFLMIN = -0.01
REFLMAX = 1.6


class _Coefficients:
    LUTS: list[np.ndarray] = []
    # resolution -> wavelength -> coefficient index
    # resolution -> band name -> coefficient index
    COEFF_INDEX_MAP: dict[int, dict[Union[tuple, str], int]] = {}

    def __init__(self, wavelength_range, resolution=0):
        self._wv_range = wavelength_range
        self._resolution = resolution

    def __call__(self):
        idx = self._find_coefficient_index(self._wv_range, resolution=self._resolution)
        band_luts = [lut_array[idx] for lut_array in self.LUTS]
        return band_luts

    def _find_coefficient_index(self, wavelength_range, resolution=0):
        """Return index in to coefficient arrays for this band's wavelength.

        This function search through the `COEFF_INDEX_MAP` dictionary and
        finds the first key where the nominal wavelength of `wavelength_range`
        falls between the minimum wavelength and maximum wavelength of the key.
        `wavelength_range` can also be the standard name of the band. For
        example, "M05" for VIIRS or "1" for MODIS.

        Args:
            wavelength_range: 3-element tuple of
                (min wavelength, nominal wavelength, max wavelength) or the
                string name of the band.
            resolution: resolution of the band to be corrected

        Returns:
            index in to coefficient arrays like `aH2O`, `aO3`, etc.
            None is returned if no matching wavelength is found

        """
        index_map = self.COEFF_INDEX_MAP
        # Find the best resolution of coefficients
        for res in sorted(index_map.keys()):
            if resolution <= res:
                index_map = index_map[res]
                break
        else:
            raise ValueError("Unrecognized data resolution: {}", resolution)
        # Find the best wavelength of coefficients
        if isinstance(wavelength_range, str):
            # wavelength range is actually a band name
            return index_map[wavelength_range]
        for lut_wvl_range, v in index_map.items():
            if isinstance(lut_wvl_range, str):
                # we are analyzing wavelengths and ignoring dataset names
                continue
            if wavelength_range[1] in lut_wvl_range:
                return v
        raise ValueError(f"Can't find LUT for {wavelength_range}.")


class _ABICoefficients(_Coefficients):
    RG_FUDGE = .55  # This number is what Ralph says "looks good" for ABI/AHI
    LUTS = [
        # aH2O
        np.array([2.4111e-003, 7.8454e-003 * RG_FUDGE, 7.9258e-3, 9.3392e-003, 2.53e-2]),
        # aO2 (bH2O for other instruments)
        np.array([1.2360e-003, 3.7296e-003, 177.7161e-006, 10.4899e-003, 1.63e-2]),
        # aO3
        np.array([4.2869e-003, 25.6509e-003 * RG_FUDGE, 802.4319e-006, 0.0000e+000, 2e-5]),
        # taur0
        np.array([184.7200e-003, 52.3490e-003, 15.8450e-003, 1.3074e-003, 311.2900e-006]),
    ]
    # resolution -> wavelength -> coefficient index
    # resolution -> band name -> coefficient index
    COEFF_INDEX_MAP = {
        2000: {
            WavelengthRange(0.450, 0.470, 0.490): 0,  # C01
            "C01": 0,
            WavelengthRange(0.590, 0.640, 0.690): 1,  # C02
            "C02": 1,
            WavelengthRange(0.8455, 0.865, 0.8845): 2,  # C03
            "C03": 2,
            # WavelengthRange((1.3705, 1.378, 1.3855)): None,  # C04 - No coefficients yet
            # "C04": None,
            WavelengthRange(1.580, 1.610, 1.640): 3,  # C05
            "C05": 3,
            WavelengthRange(2.225, 2.250, 2.275): 4,  # C06
            "C06": 4
        },
    }


class _VIIRSCoefficients(_Coefficients):
    # Values from crefl 1.7.1
    LUTS = [
        # aH2O
        np.array([0.000406601, 0.0015933, 0, 1.78644e-05, 0.00296457, 0.000617252, 0.000996563, 0.00222253, 0.00094005,
                  0.000563288, 0, 0, 0, 0, 0, 0]),
        # bH2O
        np.array([0.812659, 0.832931, 1., 0.8677850, 0.806816, 0.944958, 0.78812, 0.791204, 0.900564, 0.942907, 0, 0,
                  0, 0, 0, 0]),
        # aO3
        np.array([0.0433461, 0.0, 0.0178299, 0.0853012, 0, 0, 0, 0.0813531, 0, 0, 0.0663, 0.0836, 0.0485, 0.0395,
                  0.0119, 0.00263]),
        # taur0
        np.array([0.04350, 0.01582, 0.16176, 0.09740, 0.00369, 0.00132, 0.00033, 0.05373, 0.01561, 0.00129, 0.1131,
                  0.0994, 0.0446, 0.0416, 0.0286, 0.0155]),
    ]
    # resolution -> wavelength -> coefficient index
    # resolution -> band name -> coefficient index
    COEFF_INDEX_MAP = {
        1000: {
            WavelengthRange(0.662, 0.6720, 0.682): 0,  # M05
            "M05": 0,
            WavelengthRange(0.846, 0.8650, 0.885): 1,  # M07
            "M07": 1,
            WavelengthRange(0.478, 0.4880, 0.498): 2,  # M03
            "M03": 2,
            WavelengthRange(0.545, 0.5550, 0.565): 3,  # M04
            "M04": 3,
            WavelengthRange(1.230, 1.2400, 1.250): 4,  # M08
            "M08": 4,
            WavelengthRange(1.580, 1.6100, 1.640): 5,  # M10
            "M10": 5,
            WavelengthRange(2.225, 2.2500, 2.275): 6,  # M11
            "M11": 6,
        },
        500: {
            WavelengthRange(0.600, 0.6400, 0.680): 7,  # I01
            "I01": 7,
            WavelengthRange(0.845, 0.8650, 0.884): 8,  # I02
            "I02": 8,
            WavelengthRange(1.580, 1.6100, 1.640): 9,  # I03
            "I03": 9,
        },
    }


class _MODISCoefficients(_Coefficients):
    # Values from crefl 1.7.1
    LUTS = [
        # aH2O
        np.array([-5.60723, -5.25251, 0, 0, -6.29824, -7.70944, -3.91877, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        # bH2O
        np.array([0.820175, 0.725159, 0, 0, 0.865732, 0.966947, 0.745342, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        # aO3
        np.array([0.0715289, 0, 0.00743232, 0.089691, 0, 0, 0, 0.001, 0.00383, 0.0225, 0.0663,
                  0.0836, 0.0485, 0.0395, 0.0119, 0.00263]),
        # taur0
        np.array([0.05100, 0.01631, 0.19325, 0.09536, 0.00366, 0.00123, 0.00043, 0.3139, 0.2375, 0.1596, 0.1131,
                  0.0994, 0.0446, 0.0416, 0.0286, 0.0155]),
    ]
    # Map of pixel resolutions -> wavelength -> coefficient index
    # Map of pixel resolutions -> band name -> coefficient index
    COEFF_INDEX_MAP = {
        1000: {
            WavelengthRange(0.620, 0.6450, 0.670): 0,
            "1": 0,
            WavelengthRange(0.841, 0.8585, 0.876): 1,
            "2": 1,
            WavelengthRange(0.459, 0.4690, 0.479): 2,
            "3": 2,
            WavelengthRange(0.545, 0.5550, 0.565): 3,
            "4": 3,
            WavelengthRange(1.230, 1.2400, 1.250): 4,
            "5": 4,
            WavelengthRange(1.628, 1.6400, 1.652): 5,
            "6": 5,
            WavelengthRange(2.105, 2.1300, 2.155): 6,
            "7": 6,
        }
    }
    COEFF_INDEX_MAP[500] = COEFF_INDEX_MAP[1000]
    COEFF_INDEX_MAP[250] = COEFF_INDEX_MAP[1000]


def run_crefl(refl,
              sensor_azimuth,
              sensor_zenith,
              solar_azimuth,
              solar_zenith,
              avg_elevation=None,
              ):
    """Run main crefl algorithm.

    All input parameters are per-pixel values meaning they are the same size
    and shape as the input reflectance data, unless otherwise stated.

    :param refl: tuple of reflectance band arrays
    :param sensor_azimuth: input swath sensor azimuth angle array
    :param sensor_zenith: input swath sensor zenith angle array
    :param solar_azimuth: input swath solar azimuth angle array
    :param solar_zenith: input swath solar zenith angle array
    :param avg_elevation: average elevation (usually pre-calculated and stored in CMGDEM.hdf)

    """
    runner_cls = _runner_class_for_sensor(refl.attrs["sensor"])
    runner = runner_cls(refl)
    corr_refl = runner(sensor_azimuth, sensor_zenith, solar_azimuth, solar_zenith, avg_elevation)
    return corr_refl


class _CREFLRunner:
    def __init__(self, refl_data_arr):
        self._is_percent = refl_data_arr.attrs["units"] == "%"
        if self._is_percent:
            attrs = refl_data_arr.attrs
            refl_data_arr = refl_data_arr / 100.0
            refl_data_arr.attrs = attrs
        self._refl = refl_data_arr

    @property
    def coeffs_cls(self) -> Type[_Coefficients]:
        raise NotImplementedError()

    def __call__(self, sensor_azimuth, sensor_zenith, solar_azimuth, solar_zenith, avg_elevation):
        refl = self._refl
        height = self._height_from_avg_elevation(avg_elevation)
        coeffs_helper = self.coeffs_cls(refl.attrs["wavelength"], refl.attrs["resolution"])
        coeffs = coeffs_helper()
        mus = np.cos(np.deg2rad(solar_zenith))
        mus = mus.where(mus >= 0)
        muv = np.cos(np.deg2rad(sensor_zenith))
        phi = solar_azimuth - sensor_azimuth
        corr_refl = self._run_crefl(mus, muv, phi, solar_zenith, sensor_zenith, height, coeffs)
        if self._is_percent:
            corr_refl = corr_refl * 100.0
        return xr.DataArray(corr_refl, dims=refl.dims, coords=refl.coords, attrs=refl.attrs)

    def _run_crefl(self, mus, muv, phi, solar_zenith, sensor_zenith, height, coeffs):
        raise NotImplementedError()

    def _height_from_avg_elevation(self, avg_elevation: Optional[np.ndarray]) -> da.Array | float:
        """Get digital elevation map data for our granule with ocean fill value set to 0."""
        if avg_elevation is None:
            LOG.debug("No average elevation information provided in CREFL")
            # height = np.zeros(lon.shape, dtype=np.float64)
            height = 0.
        else:
            LOG.debug("Using average elevation information provided to CREFL")
            lon, lat = self._refl.attrs["area"].get_lonlats(chunks=self._refl.chunks)
            height = da.map_blocks(_space_mask_height, lon, lat, avg_elevation,
                                   chunks=lon.chunks, dtype=avg_elevation.dtype)
        return height


class _ABICREFLRunner(_CREFLRunner):
    @property
    def coeffs_cls(self) -> Type[_Coefficients]:
        return _ABICoefficients

    def _run_crefl(self, mus, muv, phi, solar_zenith, sensor_zenith, height, coeffs):
        LOG.debug("Using ABI CREFL algorithm")
        return da.map_blocks(_run_crefl_abi, self._refl.data, mus.data, muv.data, phi.data,
                             solar_zenith.data, sensor_zenith.data, height, *coeffs,
                             meta=np.ndarray((), dtype=self._refl.dtype),
                             chunks=self._refl.chunks, dtype=self._refl.dtype,
                             )


class _VIIRSMODISCREFLRunner(_CREFLRunner):
    def _run_crefl(self, mus, muv, phi, solar_zenith, sensor_zenith, height, coeffs):
        return da.map_blocks(_run_crefl, self._refl.data, mus.data, muv.data, phi.data,
                             height, self._refl.attrs.get("sensor"), *coeffs,
                             meta=np.ndarray((), dtype=self._refl.dtype),
                             chunks=self._refl.chunks, dtype=self._refl.dtype,
                             )


class _VIIRSCREFLRunner(_VIIRSMODISCREFLRunner):
    @property
    def coeffs_cls(self) -> Type[_Coefficients]:
        return _VIIRSCoefficients

    def _run_crefl(self, mus, muv, phi, solar_zenith, sensor_zenith, height, coeffs):
        LOG.debug("Using VIIRS CREFL algorithm")
        return super()._run_crefl(mus, muv, phi, solar_zenith, sensor_zenith, height, coeffs)


class _MODISCREFLRunner(_VIIRSMODISCREFLRunner):
    @property
    def coeffs_cls(self) -> Type[_Coefficients]:
        return _MODISCoefficients

    def _run_crefl(self, mus, muv, phi, solar_zenith, sensor_zenith, height, coeffs):
        LOG.debug("Using MODIS CREFL algorithm")
        return super()._run_crefl(mus, muv, phi, solar_zenith, sensor_zenith, height, coeffs)


_SENSOR_TO_RUNNER = {
    "abi": _ABICREFLRunner,
    "viirs": _VIIRSCREFLRunner,
    "modis": _MODISCREFLRunner,
}


def _runner_class_for_sensor(sensor_name: str) -> Type[_CREFLRunner]:
    try:
        return _SENSOR_TO_RUNNER[sensor_name]
    except KeyError:
        raise NotImplementedError(f"Don't know how to apply CREFL to data from sensor {sensor_name}.")


def _space_mask_height(lon, lat, avg_elevation):
    row = ((90.0 - lat) * avg_elevation.shape[0] / 180.0)
    col = ((lon + 180.0) * avg_elevation.shape[1] / 360.0)
    np.clip(row, 0, avg_elevation.shape[0] - 1, out=row)
    np.clip(col, 0, avg_elevation.shape[1] - 1, out=col)
    row = row.astype(np.int32)
    col = col.astype(np.int32)
    # conditions need to be this way to include NaNs
    bad_mask = ~((lon >= -180) | (lon <= 180) | (lat >= -90) | (lat <= 90))
    # convert any NaNs to valid indexes
    row[bad_mask] = 0
    col[bad_mask] = 0

    height = avg_elevation[row, col]
    # negative heights aren't allowed, clip to 0
    height[(height < 0.0) | np.isnan(height) | bad_mask] = 0.0
    return height


def _run_crefl(refl, mus, muv, phi, height, sensor_name, *coeffs):
    atm_vars_cls = _VIIRSAtmosphereVariables if sensor_name.lower() == "viirs" else _MODISAtmosphereVariables
    atm_vars = atm_vars_cls(mus, muv, phi, height, *coeffs)
    sphalb, rhoray, TtotraytH2O, tOG = atm_vars()
    return _correct_refl(refl, tOG, rhoray, TtotraytH2O, sphalb)


def _run_crefl_abi(refl, mus, muv, phi, solar_zenith, sensor_zenith, height,
                   *coeffs):
    a_O3 = [268.45, 0.5, 115.42, -3.2922]
    a_H2O = [0.0311, 0.1, 92.471, -1.3814]
    a_O2 = [0.4567, 0.007, 96.4884, -1.6970]
    G_O3 = _G_calc(solar_zenith, a_O3) + _G_calc(sensor_zenith, a_O3)
    G_H2O = _G_calc(solar_zenith, a_H2O) + _G_calc(sensor_zenith, a_H2O)
    G_O2 = _G_calc(solar_zenith, a_O2) + _G_calc(sensor_zenith, a_O2)
    # Note: bh2o values are actually ao2 values for abi
    atm_vars = _ABIAtmosphereVariables(G_O3, G_H2O, G_O2,
                                       mus, muv, phi, height, *coeffs)
    sphalb, rhoray, TtotraytH2O, tOG = atm_vars()
    return _correct_refl(refl, tOG, rhoray, TtotraytH2O, sphalb)


def _G_calc(zenith, a_coeff):
    return (np.cos(np.deg2rad(zenith))+(a_coeff[0]*(zenith**a_coeff[1])*(a_coeff[2]-zenith)**a_coeff[3]))**-1


def _correct_refl(refl, tOG, rhoray, TtotraytH2O, sphalb):
    corr_refl = (refl / tOG - rhoray) / TtotraytH2O
    corr_refl /= (1.0 + corr_refl * sphalb)
    return corr_refl.clip(REFLMIN, REFLMAX)


class _AtmosphereVariables:
    def __init__(self, mus, muv, phi, height, ah2o, bh2o, ao3, tau):
        self._mus = mus
        self._muv = muv
        self._phi = phi
        self._height = height
        self._ah2o = ah2o
        self._bh2o = bh2o
        self._ao3 = ao3
        self._tau = tau
        self._taustep4sphalb = TAUSTEP4SPHALB

    def __call__(self):
        tau_step = np.linspace(
            self._taustep4sphalb,
            MAXNUMSPHALBVALUES * self._taustep4sphalb,
            MAXNUMSPHALBVALUES)
        sphalb0 = _csalbr(tau_step)
        taur = self._tau * np.exp(-self._height / SCALEHEIGHT)
        rhoray, trdown, trup = _chand(self._phi, self._muv, self._mus, taur)
        sphalb = sphalb0[(taur / self._taustep4sphalb + 0.5).astype(np.int32)]
        Ttotrayu = ((2 / 3. + self._muv) + (2 / 3. - self._muv) * trup) / (4 / 3. + taur)
        Ttotrayd = ((2 / 3. + self._mus) + (2 / 3. - self._mus) * trdown) / (4 / 3. + taur)

        tH2O = self._get_th2o()
        TtotraytH2O = Ttotrayu * Ttotrayd * tH2O

        tO2 = self._get_to2()
        tO3 = self._get_to3()
        tOG = tO3 * tO2
        return sphalb, rhoray, TtotraytH2O, tOG

    def _get_to2(self):
        return 1.0

    def _get_to3(self):
        raise NotImplementedError()

    def _get_th2o(self):
        raise NotImplementedError()


class _ABIAtmosphereVariables(_AtmosphereVariables):
    def __init__(self, G_O3, G_H2O, G_O2, *args):
        super().__init__(*args)
        self._G_O3 = G_O3
        self._G_H2O = G_H2O
        self._G_O2 = G_O2
        self._taustep4sphalb = TAUSTEP4SPHALB_ABI

    def _get_to2(self):
        # NOTE: bh2o is actually ao2 for ABI
        return np.exp(-self._G_O2 * self._bh2o)

    def _get_to3(self):
        return np.exp(-self._G_O3 * self._ao3) if self._ao3 != 0 else 1.0

    def _get_th2o(self):
        return np.exp(-self._G_H2O * self._ah2o) if self._ah2o != 0 else 1.0


class _VIIRSAtmosphereVariables(_AtmosphereVariables):
    def __init__(self, *args):
        super().__init__(*args)
        self._airmass = self._compute_airmass()

    def _compute_airmass(self):
        air_mass = 1.0 / self._mus + 1 / self._muv
        air_mass[air_mass > MAXAIRMASS] = -1.0
        return air_mass

    def _get_to3(self):
        if self._ao3 == 0:
            return 1.0
        return np.exp(-self._airmass * UO3_VIIRS * self._ao3)

    def _get_th2o(self):
        if self._bh2o == 0:
            return 1.0
        return np.exp(-(self._ah2o * ((self._airmass * UH2O_VIIRS) ** self._bh2o)))


class _MODISAtmosphereVariables(_VIIRSAtmosphereVariables):
    def _get_to3(self):
        if self._ao3 == 0:
            return 1.0
        return np.exp(-self._airmass * UO3_MODIS * self._ao3)

    def _get_th2o(self):
        if self._bh2o == 0:
            return 1.0
        return np.exp(-np.exp(self._ah2o + self._bh2o * np.log(self._airmass * UH2O_MODIS)))


def _csalbr(tau):
    # Previously 3 functions csalbr fintexp1, fintexp3
    a = [-.57721566, 0.99999193, -0.24991055, 0.05519968, -0.00976004,
         0.00107857]
    # xx = a[0] + a[1] * tau + a[2] * tau**2 + a[3] * tau**3 + a[4] * tau**4 + a[5] * tau**5
    # xx = np.polyval(a[::-1], tau)

    # xx = a[0]
    # xftau = 1.0
    # for i in xrange(5):
    #     xftau = xftau*tau
    #     xx = xx + a[i] * xftau
    fintexp1 = np.polyval(a[::-1], tau) - np.log(tau)
    fintexp3 = (np.exp(-tau) * (1.0 - tau) + tau**2 * fintexp1) / 2.0

    return (3.0 * tau - fintexp3 *
            (4.0 + 2.0 * tau) + 2.0 * np.exp(-tau)) / (4.0 + 3.0 * tau)


def _chand(phi, muv, mus, taur):
    # FROM FUNCTION CHAND
    # phi: azimuthal difference between sun and observation in degree
    #      (phi=0 in backscattering direction)
    # mus: cosine of the sun zenith angle
    # muv: cosine of the observation zenith angle
    # taur: molecular optical depth
    # rhoray: molecular path reflectance
    # constant xdep: depolarization factor (0.0279)
    #          xfd = (1-xdep/(2-xdep)) / (1 + 2*xdep/(2-xdep)) = 2 * (1 - xdep) / (2 + xdep) = 0.958725775
    # */
    xfd = 0.958725775
    xbeta2 = 0.5
    #         float pl[5];
    #         double fs01, fs02, fs0, fs1, fs2;
    as0 = [0.33243832, 0.16285370, -0.30924818, -0.10324388, 0.11493334,
           -6.777104e-02, 1.577425e-03, -1.240906e-02, 3.241678e-02,
           -3.503695e-02]
    as1 = [0.19666292, -5.439061e-02]
    as2 = [0.14545937, -2.910845e-02]
    #         float phios, xcos1, xcos2, xcos3;
    #         float xph1, xph2, xph3, xitm1, xitm2;
    #         float xlntaur, xitot1, xitot2, xitot3;
    #         int i, ib;

    xph1 = 1.0 + (3.0 * mus * mus - 1.0) * (3.0 * muv * muv - 1.0) * xfd / 8.0
    xph2 = -xfd * xbeta2 * 1.5 * mus * muv * np.sqrt(
        1.0 - mus * mus) * np.sqrt(1.0 - muv * muv)
    xph3 = xfd * xbeta2 * 0.375 * (1.0 - mus * mus) * (1.0 - muv * muv)

    # pl[0] = 1.0
    # pl[1] = mus + muv
    # pl[2] = mus * muv
    # pl[3] = mus * mus + muv * muv
    # pl[4] = mus * mus * muv * muv

    fs01 = as0[0] + (mus + muv) * as0[1] + (mus * muv) * as0[2] + (
            mus * mus + muv * muv) * as0[3] + (mus * mus * muv * muv) * as0[4]
    fs02 = as0[5] + (mus + muv) * as0[6] + (mus * muv) * as0[7] + (
            mus * mus + muv * muv) * as0[8] + (mus * mus * muv * muv) * as0[9]
    #         for (i = 0; i < 5; i++) {
    #                 fs01 += (double) (pl[i] * as0[i]);
    #                 fs02 += (double) (pl[i] * as0[5 + i]);
    #         }

    # for refl, (ah2o, bh2o, ao3, tau) in zip(reflectance_bands, coefficients):

    # ib = _find_coefficient_index(center_wl)
    # if ib is None:
    #     raise ValueError("Can't handle band with wavelength '{}'".format(center_wl))

    xlntaur = np.log(taur)

    fs0 = fs01 + fs02 * xlntaur
    fs1 = as1[0] + xlntaur * as1[1]
    fs2 = as2[0] + xlntaur * as2[1]
    del xlntaur, fs01, fs02

    trdown = np.exp(-taur / mus)
    trup = np.exp(-taur / muv)

    xitm1 = (1.0 - trdown * trup) / 4.0 / (mus + muv)
    xitm2 = (1.0 - trdown) * (1.0 - trup)
    xitot1 = xph1 * (xitm1 + xitm2 * fs0)
    xitot2 = xph2 * (xitm1 + xitm2 * fs1)
    xitot3 = xph3 * (xitm1 + xitm2 * fs2)
    del xph1, xph2, xph3, xitm1, xitm2, fs0, fs1, fs2

    phios = np.deg2rad(phi + 180.0)
    xcos1 = 1.0
    xcos2 = np.cos(phios)
    xcos3 = np.cos(2.0 * phios)
    del phios

    rhoray = xitot1 * xcos1 + xitot2 * xcos2 * 2.0 + xitot3 * xcos3 * 2.0
    return rhoray, trdown, trup



# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""CF processing of attributes."""
from __future__ import annotations

import datetime as dt
import json
import logging
from collections import OrderedDict

import numpy as np
import xarray as xr

from satpy.utils import flatten_dict

logger = logging.getLogger(__name__)


class AttributeEncoder(json.JSONEncoder):
    """JSON encoder for dataset attributes."""

    def default(self, obj):
        """Return a json-serializable object for *obj*.

        In order to facilitate decoding, elements in dictionaries, lists/tuples and multi-dimensional arrays are
        encoded recursively.
        """
        if isinstance(obj, dict):
            serialized = {}
            for key, val in obj.items():
                serialized[key] = self.default(val)
            return serialized
        elif isinstance(obj, (list, tuple, np.ndarray)):
            return [self.default(item) for item in obj]
        return self._encode(obj)

    def _encode(self, obj):
        """Encode the given object as a json-serializable datatype."""
        if isinstance(obj, (bool, np.bool_)):
            # Bool has to be checked first, because it is a subclass of int
            return str(obj).lower()
        elif isinstance(obj, (int, float, str)):
            return obj
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.void):
            return tuple(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return str(obj)


def _encode_numpy_array(obj):
    """Encode numpy array as a netCDF4 serializable datatype."""
    from satpy.writers.cf_writer import NC4_DTYPES

    # Only plain 1-d arrays are supported. Skip record arrays and multi-dimensional arrays.
    is_plain_1d = not obj.dtype.fields and len(obj.shape) <= 1
    if not is_plain_1d:
        raise ValueError("Only a 1D numpy array can be encoded as netCDF attribute.")
    if obj.dtype in NC4_DTYPES:
        return obj
    if obj.dtype == np.bool_:
        # Boolean arrays are not supported, convert to array of strings.
        return [s.lower() for s in obj.astype(str)]
    return obj.tolist()


def _encode_object(obj):
    """Try to encode `obj` as a netCDF/Zarr compatible datatype which most closely resembles the object's nature.

    Raises:
        ValueError: if no such datatype could be found
    """
    is_nonbool_int = isinstance(obj, int) and not isinstance(obj, (bool, np.bool_))
    is_encode_type = isinstance(obj, (float, str, np.integer, np.floating))
    if is_nonbool_int or is_encode_type:
        return obj
    elif isinstance(obj, np.ndarray):
        return _encode_numpy_array(obj)
    raise ValueError("Unable to encode")


def _try_decode_object(obj):
    """Try to decode byte string."""
    try:
        decoded = obj.decode()
    except AttributeError:
        decoded = obj
    return decoded


def _encode_python_objects(obj):
    """Try to find the datatype which most closely resembles the object's nature.

    If on failure, encode as a string. Plain lists are encoded recursively.
    """
    if isinstance(obj, (list, tuple)) and all([not isinstance(item, (list, tuple)) for item in obj]):
        return [_encode_to_cf(item) for item in obj]
    try:
        dump = _encode_object(obj)
    except ValueError:
        decoded = _try_decode_object(obj)
        dump = json.dumps(decoded, cls=AttributeEncoder).strip('"')
    return dump


def _encode_to_cf(obj):
    """Encode the given object as a netcdf compatible datatype."""
    try:
        return obj.to_cf()
    except AttributeError:
        return _encode_python_objects(obj)


def encode_attrs_to_cf(attrs):
    """Encode dataset attributes as a netcdf compatible datatype.

    Args:
        attrs (dict):
            Attributes to be encoded
    Returns:
        dict: Encoded (and sorted) attributes

    """
    encoded_attrs = []
    for key, val in sorted(attrs.items()):
        if val is not None:
            encoded_attrs.append((key, _encode_to_cf(val)))
    return OrderedDict(encoded_attrs)


def preprocess_attrs(
        data_arr: xr.DataArray,
        flatten_attrs: bool,
        exclude_attrs: list[str] | None
) -> xr.DataArray:
    """Preprocess DataArray attributes to be written into CF-compliant netCDF/Zarr."""
    _drop_attrs(data_arr, exclude_attrs)
    _add_ancillary_variables_attrs(data_arr)
    _format_prerequisites_attrs(data_arr)

    if "long_name" not in data_arr.attrs and "standard_name" not in data_arr.attrs:
        data_arr.attrs["long_name"] = data_arr.name

    if flatten_attrs:
        data_arr.attrs = flatten_dict(data_arr.attrs)

    data_arr.attrs = encode_attrs_to_cf(data_arr.attrs)

    return data_arr


def _drop_attrs(
        data_arr: xr.DataArray,
        user_excluded_attrs: list[str] | None
) -> None:
    """Remove undesirable attributes."""
    attrs_to_drop = (
            (user_excluded_attrs or []) +
            _get_satpy_attrs(data_arr) +
            _get_none_attrs(data_arr) +
            ["area"]
    )
    for key in attrs_to_drop:
        data_arr.attrs.pop(key, None)


def _get_satpy_attrs(data_arr: xr.DataArray) -> list[str]:
    """Remove _satpy attribute."""
    return [key for key in data_arr.attrs if key.startswith("_satpy")] + ["_last_resampler"]


def _get_none_attrs(data_arr: xr.DataArray) -> list[str]:
    """Remove attribute keys with None value."""
    return [attr_name for attr_name, attr_val in data_arr.attrs.items() if attr_val is None]


def _add_ancillary_variables_attrs(data_arr: xr.DataArray) -> None:
    """Replace ancillary_variables DataArray with a list of their name."""
    list_ancillary_variable_names = [da_ancillary.attrs["name"]
                                     for da_ancillary in data_arr.attrs.get("ancillary_variables", [])]
    if list_ancillary_variable_names:
        data_arr.attrs["ancillary_variables"] = " ".join(list_ancillary_variable_names)
    else:
        data_arr.attrs.pop("ancillary_variables", None)


def _format_prerequisites_attrs(data_arr: xr.DataArray) -> None:
    """Reformat prerequisites attribute value to string."""
    if "prerequisites" in data_arr.attrs:
        data_arr.attrs["prerequisites"] = [np.bytes_(str(prereq)) for prereq in data_arr.attrs["prerequisites"]]


def _add_history(attrs):
    """Add 'history' attribute to dictionary."""
    _history_create = "Created by pytroll/satpy on {}".format(dt.datetime.now(dt.timezone.utc))
    if "history" in attrs:
        if isinstance(attrs["history"], list):
            attrs["history"] = "".join(attrs["history"])
        attrs["history"] += "\n" + _history_create
    else:
        attrs["history"] = _history_create
    return attrs


def preprocess_header_attrs(header_attrs, flatten_attrs=False):
    """Prepare file header attributes."""
    if header_attrs is not None:
        if flatten_attrs:
            header_attrs = flatten_dict(header_attrs)
        header_attrs = encode_attrs_to_cf(header_attrs)  # OrderedDict
    else:
        header_attrs = {}
    header_attrs = _add_history(header_attrs)
    return header_attrs



# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Simple decision tree functionality."""
from __future__ import annotations

from typing import Callable

from satpy.utils import get_logger, recursive_dict_update

LOG = get_logger(__name__)


class DecisionTree:
    """Structure to search for nearest match from a set of parameters.

    This class is used to find the best configuration section by matching
    a set of attributes. The provided dictionary contains a mapping of
    "section name" to "decision" dictionaries. Each decision dictionary
    contains the attributes that will be used for matching plus any
    additional keys that could be useful when matched. This class will
    search these decisions and return the one with the most matching
    parameters to the attributes passed to the
    :meth:`~DecisionTree.find_match` method.

    Note that decision sections are provided as a dict instead of a list
    so that they can be overwritten or updated by doing the equivalent
    of a ``current_dicts.update(new_dicts)``.

    Examples:
        Decision sections are provided as a dictionary of dictionaries.
        The returned match will be the first result found by searching
        provided `match_keys` in order.

        ::

            decisions = {
                'first_section': {
                    'a': 1,
                    'b': 2,
                    'useful_key': 'useful_value',
                },
                'second_section': {
                    'a': 5,
                    'useful_key': 'other_useful_value1',
                },
                'third_section': {
                    'b': 4,
                    'useful_key': 'other_useful_value2',
                },
            }
            tree = DecisionTree(decisions, ('a', 'b'))
            tree.find_match(a=5, b=2)  # second_section dict
            tree.find_match(a=1, b=2)  # first_section dict
            tree.find_match(a=5, b=4)  # second_section dict
            tree.find_match(a=3, b=2)  # no match

    """

    any_key = None
    _indent = "  "

    def __init__(self, decision_dicts, match_keys, multival_keys=None):
        """Init the decision tree.

        Args:
            decision_dicts (dict): Dictionary of dictionaries. Each
                sub-dictionary contains key/value pairs that can be
                matched from the `find_match` method. Sub-dictionaries
                can include additional keys outside the ``match_keys``
                provided to act as the "result" of a query. The keys of
                the root dict are arbitrary.
            match_keys (list): Keys of the provided dictionary to use for
                matching.
            multival_keys (list): Keys of `match_keys` that can be provided
                as multiple values.
                A multi-value key can be specified as a single value
                (typically a string) or a set. If a set, it will be sorted
                and converted to a tuple and then used for matching.
                When querying the tree, these keys will
                be searched for exact multi-value results (the sorted tuple)
                and if not found then each of the values will be searched
                individually in alphabetical order.

        """
        self._match_keys = match_keys
        self._multival_keys = multival_keys or []
        self._log = get_logger(__name__ + f".{self.__class__.__name__}")
        self._tree = _DecisionDict(self._match_keys[0], 0)
        if not isinstance(decision_dicts, (list, tuple)):
            decision_dicts = [decision_dicts]
        self.add_config_to_tree(*decision_dicts)

    def _indented_trace(self, indent_level: int, msg: str):
        indent = self._indent * indent_level
        self._log.trace(f"{indent}{msg}")

    def _indented_print(self, indent_level: int, msg: str):
        indent = self._indent * indent_level
        print(f"{indent}{msg}")  # noqa: T201

    def print_tree(self):
        """Print the decision tree in a structured human-readable format."""
        self._print_tree_level(0, self._tree)

    def _print_tree_level(self, level: int, curr_level: dict):
        if len(self._match_keys) == level:
            # final component
            self._print_matched_info(len(self._match_keys), curr_level, self._indented_print)
            return

        match_key = self._match_keys[level]
        for match_val, next_level in curr_level.items():
            if match_val is None:
                match_val = "<wildcard>"
            self._indented_print(level, f"{match_key}={match_val}")
            self._print_tree_level(level + 1, next_level)

    def _print_matched_info(self, level: int, decision_info: dict, print_func: Callable) -> None:
        any_keys = False
        for key in self._match_keys:
            if key not in decision_info:
                continue
            any_keys = True
            print_func(level, f"| {key}={decision_info[key]}")
        if not any_keys:
            print_func(level, "| <global wildcard match>")

    def add_config_to_tree(self, *decision_dicts):
        """Add a configuration to the tree."""
        conf = {}
        for decision_dict in decision_dicts:
            recursive_dict_update(conf, decision_dict)
        self._build_tree(conf)

    def _build_tree(self, conf):
        """Build the tree.

        Create a tree structure of dicts where each level represents the
        possible matches for a specific ``match_key``. When finding matches
        we will iterate through the tree matching each key that we know about.
        The last dict in the "tree" will contain the configure section whose
        match values led down that path in the tree.

        See :meth:`DecisionTree.find_match` for more information.

        """
        for _section_name, sect_attrs in conf.items():
            # Set a path in the tree for each section in the config files
            curr_level = self._tree
            for match_level, match_key in enumerate(self._match_keys):
                # or None is necessary if they have empty strings
                this_attr_val = sect_attrs.get(match_key, self.any_key) or None
                if match_key in self._multival_keys and isinstance(this_attr_val, list):
                    this_attr_val = tuple(sorted(this_attr_val))
                is_last_key = match_key == self._match_keys[-1]
                level_needs_init = this_attr_val not in curr_level
                if is_last_key:
                    # if we are at the last attribute, then assign the value
                    # set the dictionary of attributes because the config is
                    # not persistent
                    curr_level[this_attr_val] = sect_attrs
                elif level_needs_init:
                    curr_level[this_attr_val] = _DecisionDict(self._match_keys[match_level + 1], match_level + 1)
                curr_level = curr_level[this_attr_val]

    @staticmethod
    def _convert_query_val_to_hashable(query_val):
        _sorted_query_val = sorted(query_val)
        query_vals = [tuple(_sorted_query_val)] + _sorted_query_val
        query_vals += query_val
        return query_vals

    def _get_query_values(self, query_dict, curr_match_key):
        query_val = query_dict[curr_match_key]
        if curr_match_key in self._multival_keys and isinstance(query_val, set):
            query_vals = self._convert_query_val_to_hashable(query_val)
        else:
            query_vals = [query_val]
        return query_vals

    def _find_match_if_known(self, curr_level, remaining_match_keys, query_dict):
        match = None
        curr_match_key = remaining_match_keys[0]
        if curr_match_key not in query_dict:
            self._indented_trace(
                len(self._match_keys) - len(remaining_match_keys),
                f"Match key {curr_match_key!r} not in query dict"
            )
            return match

        query_vals = self._get_query_values(query_dict, curr_match_key)
        for query_val in query_vals:
            if not curr_level.traced_contains(query_val):
                continue
            match = self._find_match(curr_level[query_val],
                                     remaining_match_keys[1:],
                                     query_dict)
            if match is not None:
                break
        return match

    def _find_match(self, curr_level, remaining_match_keys, query_dict):
        """Find a match."""
        if len(remaining_match_keys) == 0:
            # we're at the bottom level, we must have found something
            self._indented_trace(len(self._match_keys), "Found match!")
            self._print_matched_info(
                len(self._match_keys),
                curr_level,
                self._indented_trace,
            )
            return curr_level

        match = self._find_match_if_known(
            curr_level, remaining_match_keys, query_dict)

        if match is None and curr_level.traced_contains(self.any_key):
            # if we couldn't find it using the attribute then continue with
            # the other attributes down the 'any' path
            match = self._find_match(
                curr_level[self.any_key],
                remaining_match_keys[1:],
                query_dict)
        return match

    def find_match(self, **query_dict):
        """Find a match.

        Recursively search through the tree structure for a path that matches
        the provided match parameters.

        """
        try:
            match = self._find_match(self._tree, self._match_keys, query_dict)
        except (KeyError, IndexError, ValueError, TypeError):
            LOG.debug("Match exception:", exc_info=True)
            LOG.error("Error when finding matching decision section")
            match = None

        if match is None:
            # only possible if no default section was provided
            raise KeyError("No decision section found for %s" %
                           (query_dict.get("uid", None),))
        return match


class _DecisionDict(dict):
    """Helper class for debugging decision tree choices.

    At the time of writing this class does not do anything extra to the
    behavior of what choices are made. It is only a record keeper to make
    log messages and debugging operations easier. A simple `dict` should
    be useable in its place.

    """
    def __init__(self, match_key: str, level: int):
        self.match_key = match_key
        self.level = level
        self._log = get_logger(__name__ + f".{self.__class__.__name__}.{match_key}")
        self._indent = DecisionTree._indent * self.level
        super().__init__()

    def _log_trace(self, msg: str):
        self._log.trace(f"{self._indent}{msg}")

    def traced_contains(self, item: str) -> bool:
        contains = super().__contains__(item)
        item_str = "<wildcard>" if item is DecisionTree.any_key else repr(item)
        self._log_trace(f"Checking {self.match_key!r} level for {item_str}: {contains}")
        return contains



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2024 Satpy developers
#
# This file is part of Satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for CAMEL Level 3 emissivity files in netCDF4 format.

For more information about the data, see: <https://lpdaac.usgs.gov/products/cam5k30emv002/>.

NOTE: This reader only supports the global 0.05 degree grid data.
"""


import datetime as dt
import logging

import xarray as xr
from pyresample import geometry

from satpy.readers.core.file_handlers import BaseFileHandler

logger = logging.getLogger(__name__)

# Area extent for the CAMEL product (global)
GLOB_AREA_EXT = [-180, -90, 180, 90]


class CAMELL3NCFileHandler(BaseFileHandler):
    """File handler for CAMEL data in netCDF format."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super().__init__(filename, filename_info, filetype_info)
        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=True,
                                  chunks={"xc": "auto", "yc": "auto"})

        if "0.05" not in self.nc.attrs["geospatial_lon_resolution"]:
            raise ValueError("Only 0.05 degree grid data is supported.")
        if "0.05" not in self.nc.attrs["geospatial_lat_resolution"]:
            raise ValueError("Only 0.05 degree grid data is supported.")

        self.nlines = self.nc.sizes["latitude"]
        self.ncols = self.nc.sizes["longitude"]
        self.area = None


    @property
    def start_time(self):
        """Start timestamp of the dataset."""
        date_str = self.nc.attrs["time_coverage_start"]
        return dt.datetime.strptime(date_str, "%Y-%m-%d %H:%M:%SZ")

    @property
    def end_time(self):
        """End timestamp of the dataset."""
        date_str = self.nc.attrs["time_coverage_end"]
        return dt.datetime.strptime(date_str, "%Y-%m-%d %H:%M:%SZ")


    def get_dataset(self, key, info):
        """Load a dataset."""
        var = info["file_key"]
        logger.debug("Reading in get_dataset %s.", var)
        variable = self.nc[var]

        # For the emissivity there are multiple bands, so we need to select the correct one
        if var == "camel_emis":
            if info["band_id"] >= variable.shape[2]:
                raise ValueError("Band id requested is larger than dataset.")
            variable = variable[:, :, info["band_id"]]

        # Rename the latitude and longitude dimensions to x and y
        variable = variable.rename({"latitude": "y", "longitude": "x"})

        variable.attrs.update(key.to_dict())
        return variable


    def get_area_def(self, dsid):
        """Get the area definition, a global lat/lon area for this type of dataset."""
        proj_param = "EPSG:4326"

        area = geometry.AreaDefinition("gridded_camel",
                                       "A global gridded area",
                                       "longlat",
                                       proj_param,
                                       self.ncols,
                                       self.nlines,
                                       GLOB_AREA_EXT)
        self.area = area
        return area



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the NinJoTIFF writer."""

import sys
import unittest
from unittest import mock

import numpy as np
import pytest
import xarray as xr


class FakeImage:
    """Fake image."""

    def __init__(self, data, mode):
        """Init fake image."""
        self.data = data
        self.mode = mode

    def get_scaling_from_history(self):
        """Return dummy scale and offset."""
        return xr.DataArray(1), xr.DataArray(0)


pyninjotiff_mock = mock.Mock()
pyninjotiff_mock.ninjotiff = mock.Mock()


@mock.patch.dict(sys.modules, {"pyninjotiff": pyninjotiff_mock, "pyninjotiff.ninjotiff": pyninjotiff_mock.ninjotiff})
class TestNinjoTIFFWriter(unittest.TestCase):
    """The ninjo tiff writer tests."""

    @mock.patch("satpy.writers.ninjotiff.nt", pyninjotiff_mock.ninjotiff)
    def test_init(self):
        """Test the init."""
        from satpy.writers.ninjotiff import NinjoTIFFWriter
        ninjo_tags = {40000: "NINJO"}
        ntw = NinjoTIFFWriter(tags=ninjo_tags)
        assert ntw.tags == ninjo_tags

    @mock.patch("satpy.writers.ninjotiff.ImageWriter.save_dataset")
    @mock.patch("satpy.writers.ninjotiff.nt", pyninjotiff_mock.ninjotiff)
    def test_dataset(self, iwsd):
        """Test saving a dataset."""
        from satpy.writers.ninjotiff import NinjoTIFFWriter
        ntw = NinjoTIFFWriter()
        dataset = xr.DataArray([1, 2, 3], attrs={"units": "K"})
        with mock.patch("satpy.writers.ninjotiff.convert_units") as uconv:
            ntw.save_dataset(dataset, physic_unit="CELSIUS")
            uconv.assert_called_once_with(dataset, "K", "CELSIUS")
        assert iwsd.call_count == 1

    @mock.patch("satpy.writers.ninjotiff.ImageWriter.save_dataset")
    @mock.patch("satpy.writers.ninjotiff.nt", pyninjotiff_mock.ninjotiff)
    def test_dataset_skip_unit_conversion(self, iwsd):
        """Test saving a dataset without unit conversion."""
        from satpy.writers.ninjotiff import NinjoTIFFWriter
        ntw = NinjoTIFFWriter()
        dataset = xr.DataArray([1, 2, 3], attrs={"units": "K"})
        with mock.patch("satpy.writers.ninjotiff.convert_units") as uconv:
            ntw.save_dataset(dataset, physic_unit="CELSIUS",
                             convert_temperature_units=False)
            uconv.assert_not_called()
        assert iwsd.call_count == 1

    @mock.patch("satpy.writers.ninjotiff.NinjoTIFFWriter.save_dataset")
    @mock.patch("satpy.writers.ninjotiff.ImageWriter.save_image")
    @mock.patch("satpy.writers.ninjotiff.nt", pyninjotiff_mock.ninjotiff)
    def test_image(self, iwsi, save_dataset):
        """Test saving an image."""
        nt = pyninjotiff_mock.ninjotiff
        nt.reset_mock()
        from satpy.writers.ninjotiff import NinjoTIFFWriter
        ntw = NinjoTIFFWriter()
        dataset = xr.DataArray([1, 2, 3], attrs={"units": "K"})
        img = FakeImage(dataset, "L")
        ret = ntw.save_image(img, filename="bla.tif", compute=False)
        nt.save.assert_called()
        assert nt.save.mock_calls[0][2]["compute"] is False
        assert nt.save.mock_calls[0][2]["ch_min_measurement_unit"] < nt.save.mock_calls[0][2]["ch_max_measurement_unit"]
        assert ret == [nt.save.return_value]

    def test_convert_units_self(self):
        """Test that unit conversion to themselves do nothing."""
        from satpy.tests.utils import make_fake_scene
        from satpy.writers.ninjotiff import convert_units

        # ensure that converting from % to itself does not change the data
        sc = make_fake_scene(
                {"VIS006": np.arange(25, dtype="f4").reshape(5, 5)},
                common_attrs={"units": "%"})
        ds_in = sc["VIS006"]
        ds_out = convert_units(ds_in, "%", "%")
        np.testing.assert_array_equal(ds_in, ds_out)
        assert ds_in.attrs == ds_out.attrs

    def test_convert_units_temp(self):
        """Test that temperature unit conversions works as expected."""
        # test converting between °C and K
        from satpy.tests.utils import make_fake_scene
        from satpy.writers.ninjotiff import convert_units
        sc = make_fake_scene(
                {"IR108": np.arange(25, dtype="f4").reshape(5, 5)},
                common_attrs={"units": "K"})
        ds_in_k = sc["IR108"]
        for out_unit in ("C", "CELSIUS"):
            ds_out_c = convert_units(ds_in_k, "K", out_unit)
            np.testing.assert_array_almost_equal(ds_in_k - 273.15, ds_out_c)
            assert ds_in_k.attrs != ds_out_c.attrs
            assert ds_out_c.attrs["units"] == out_unit
        # test that keys aren't lost
        assert ds_out_c.attrs.keys() - ds_in_k.attrs.keys() <= {"units"}
        assert ds_in_k.attrs.keys() <= ds_out_c.attrs.keys()

    def test_convert_units_other(self):
        """Test that other unit conversions are not implemented."""
        # test arbitrary different conversion
        from satpy.tests.utils import make_fake_scene
        from satpy.writers.ninjotiff import convert_units
        sc = make_fake_scene(
                {"rain_rate": np.arange(25, dtype="f8").reshape(5, 5)},
                common_attrs={"units": "millimeter/hour"})

        ds_in = sc["rain_rate"]
        with pytest.raises(NotImplementedError):
            convert_units(ds_in, "millimeter/hour", "m/s")

    @mock.patch("satpy.writers.ninjotiff.NinjoTIFFWriter.save_dataset")
    @mock.patch("satpy.writers.ninjotiff.ImageWriter.save_image")
    @mock.patch("satpy.writers.ninjotiff.nt", pyninjotiff_mock.ninjotiff)
    def test_P_image_is_uint8(self, iwsi, save_dataset):
        """Test that a P-mode image is converted to uint8s."""
        nt = pyninjotiff_mock.ninjotiff
        nt.reset_mock()
        from satpy.writers.ninjotiff import NinjoTIFFWriter
        ntw = NinjoTIFFWriter()
        dataset = xr.DataArray([1, 2, 3]).astype(int)
        img = FakeImage(dataset, "P")
        ntw.save_image(img, filename="bla.tif", compute=False)
        assert nt.save.mock_calls[0][1][0].data.dtype == np.uint8



#!/usr/bin/env python
# Copyright (c) 2018-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for compositor config handling."""

import satpy


def test_bad_sensor_yaml_configs(tmp_path):
    """Test composite YAML file with no sensor isn't loaded.

    But the bad YAML also shouldn't crash composite configuration loading.

    """
    from satpy.composites.config_loader import load_compositor_configs_for_sensors

    comp_dir = tmp_path / "composites"
    comp_dir.mkdir()
    comp_yaml = comp_dir / "fake_sensor.yaml"
    with satpy.config.set(config_path=[tmp_path]):
        _create_fake_composite_config(comp_yaml)

        # no sensor_name in YAML, quietly ignored
        comps, _ = load_compositor_configs_for_sensors(["fake_sensor"])
        assert "fake_sensor" in comps
        assert "fake_composite" not in comps["fake_sensor"]


def _create_fake_composite_config(yaml_filename: str):
    import yaml

    from satpy.composites.aux_data import StaticImageCompositor

    with open(yaml_filename, "w") as comp_file:
        yaml.dump({
            "composites": {
                "fake_composite": {
                    "compositor": StaticImageCompositor,
                    "url": "http://example.com/image.png",
                },
            },
        },
            comp_file,
        )



# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the enhancements overlays module."""
from __future__ import annotations

import warnings
from unittest import mock

import xarray as xr
from dask import array as da
from trollimage.colormap import greys


class TestOverlays:
    """Tests for add_overlay and add_decorate functions."""

    def setup_method(self):
        """Create test data and mock pycoast/pydecorate."""
        from pyresample.geometry import AreaDefinition
        from trollimage.xrimage import XRImage

        proj_dict = {"proj": "lcc", "datum": "WGS84", "ellps": "WGS84",
                     "lon_0": -95., "lat_0": 25, "lat_1": 25,
                     "units": "m", "no_defs": True}
        self.area_def = AreaDefinition(
            "test", "test", "test", proj_dict,
            200, 400, (-1000., -1500., 1000., 1500.),
        )
        self.orig_rgb_img = XRImage(
            xr.DataArray(da.arange(75., chunks=10).reshape(3, 5, 5) / 75.,
                         dims=("bands", "y", "x"),
                         coords={"bands": ["R", "G", "B"]},
                         attrs={"name": "test_ds", "area": self.area_def})
        )
        self.orig_l_img = XRImage(
            xr.DataArray(da.arange(25., chunks=10).reshape(5, 5) / 75.,
                         dims=("y", "x"),
                         attrs={"name": "test_ds", "area": self.area_def})
        )

        self.decorate = {
            "decorate": [
                {"logo": {"logo_path": "", "height": 143, "bg": "white", "bg_opacity": 255}},
                {"text": {
                    "txt": "TEST",
                    "align": {"top_bottom": "bottom", "left_right": "right"},
                    "font": "",
                    "font_size": 22,
                    "height": 30,
                    "bg": "black",
                    "bg_opacity": 255,
                    "line": "white"}},
                {"scale": {
                    "colormap": greys,
                    "extend": False,
                    "width": 1670, "height": 110,
                    "tick_marks": 5, "minor_tick_marks": 1,
                    "cursor": [0, 0], "bg": "white",
                    "title": "TEST TITLE OF SCALE",
                    "fontsize": 110, "align": "cc"
                }}
            ]
        }

        import_mock = mock.MagicMock()
        modules = {"pycoast": import_mock.pycoast,
                   "pydecorate": import_mock.pydecorate}
        self.module_patcher = mock.patch.dict("sys.modules", modules)
        self.module_patcher.start()

    def teardown_method(self):
        """Turn off pycoast/pydecorate mocking."""
        self.module_patcher.stop()

    def test_add_overlay_basic_rgb(self):
        """Test basic add_overlay usage with RGB data."""
        from pycoast import ContourWriterAGG

        from satpy.enhancements.overlays import _burn_overlay, add_overlay
        coast_dir = "/path/to/coast/data"
        with mock.patch.object(self.orig_rgb_img, "apply_pil") as apply_pil:
            apply_pil.return_value = self.orig_rgb_img
            new_img = add_overlay(self.orig_rgb_img, self.area_def, coast_dir, fill_value=0)
            assert self.orig_rgb_img.mode == new_img.mode
            new_img = add_overlay(self.orig_rgb_img, self.area_def, coast_dir)
            assert self.orig_rgb_img.mode + "A" == new_img.mode

            with mock.patch.object(self.orig_rgb_img, "convert") as convert:
                convert.return_value = self.orig_rgb_img
                overlays = {"coasts": {"outline": "red"}}
                new_img = add_overlay(self.orig_rgb_img, self.area_def, coast_dir,
                                      overlays=overlays, fill_value=0)
                pil_args = None
                pil_kwargs = {"fill_value": 0}
                fun_args = (self.orig_rgb_img.data.area, ContourWriterAGG.return_value, overlays)
                fun_kwargs = None
                apply_pil.assert_called_with(_burn_overlay, self.orig_rgb_img.mode,
                                             pil_args, pil_kwargs, fun_args, fun_kwargs)
                ContourWriterAGG.assert_called_with(coast_dir)

                # test legacy call

                grid = {"minor_is_tick": True}
                color = "red"
                expected_overlays = {"coasts": {"outline": color, "width": 0.5, "level": 1},
                                     "borders": {"outline": color, "width": 0.5, "level": 1},
                                     "grid": grid}
                with warnings.catch_warnings(record=True) as wns:
                    warnings.simplefilter("always")
                    new_img = add_overlay(self.orig_rgb_img, self.area_def, coast_dir,
                                          color=color, grid=grid, fill_value=0)
                    assert len(wns) == 1
                    assert issubclass(wns[0].category, DeprecationWarning)
                    assert "deprecated" in str(wns[0].message)

                pil_args = None
                pil_kwargs = {"fill_value": 0}
                fun_args = (self.orig_rgb_img.data.area, ContourWriterAGG.return_value, expected_overlays)
                fun_kwargs = None
                apply_pil.assert_called_with(_burn_overlay, self.orig_rgb_img.mode,
                                             pil_args, pil_kwargs, fun_args, fun_kwargs)
                ContourWriterAGG.assert_called_with(coast_dir)

    def test_add_overlay_basic_l(self):
        """Test basic add_overlay usage with L data."""
        from satpy.enhancements.overlays import add_overlay
        new_img = add_overlay(self.orig_l_img, self.area_def, "", fill_value=0)
        assert "RGB" == new_img.mode
        new_img = add_overlay(self.orig_l_img, self.area_def, "")
        assert "RGBA" == new_img.mode

    def test_add_decorate_basic_rgb(self):
        """Test basic add_decorate usage with RGB data."""
        from satpy.enhancements.overlays import add_decorate
        new_img = add_decorate(self.orig_rgb_img, **self.decorate)
        assert "RGBA" == new_img.mode

    def test_add_decorate_basic_l(self):
        """Test basic add_decorate usage with L data."""
        from satpy.enhancements.overlays import add_decorate
        new_img = add_decorate(self.orig_l_img, **self.decorate)
        assert "RGBA" == new_img.mode



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to CLAVR-X HDF4 products."""

from __future__ import annotations

import logging
import os
from glob import glob
from typing import Optional

import netCDF4
import numpy as np
import xarray as xr
from pyresample import geometry

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.hdf4 import SDS, HDF4FileHandler
from satpy.utils import get_legacy_chunk_size

LOG = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

CF_UNITS = {
    "none": "1",
}

SENSORS = {
    "MODIS": "modis",
    "VIIRS": "viirs",
    "AVHRR": "avhrr",
    "AHI": "ahi",
    "ABI": "abi",
    "GOES-RU-IMAGER": "abi",
}
PLATFORMS = {
    "SNPP": "npp",
    "HIM8": "himawari8",
    "HIM9": "himawari9",
    "H08": "himawari8",
    "H09": "himawari9",
    "G16": "GOES-16",
    "G17": "GOES-17",
    "G18": "GOES-18",
}
ROWS_PER_SCAN = {
    "viirs": 16,
    "modis": 10,
}
NADIR_RESOLUTION = {
    "viirs": 742,
    "modis": 1000,
    "avhrr": 1050,
    "ahi": 2000,
    "abi": 2004,
}

CHANNEL_ALIASES = {
    "abi": {"refl_0_47um_nom": {"name": "C01", "wavelength": 0.47, "modifiers": ("sunz_corrected",)},
            "refl_0_65um_nom": {"name": "C02", "wavelength": 0.64, "modifiers": ("sunz_corrected",)},
            "refl_0_86um_nom": {"name": "C03", "wavelength": 0.865, "modifiers": ("sunz_corrected",)},
            "refl_1_38um_nom": {"name": "C04", "wavelength": 1.38, "modifiers": ("sunz_corrected",)},
            "refl_1_60um_nom": {"name": "C05", "wavelength": 1.61, "modifiers": ("sunz_corrected",)},
            "refl_2_10um_nom": {"name": "C06", "wavelength": 2.25, "modifiers": ("sunz_corrected",)},
            },
    "viirs": {"refl_0_65um_nom": {"name": "I01", "wavelength": 0.64, "modifiers": ("sunz_corrected",)},
              "refl_1_38um_nom": {"name": "M09", "wavelength": 1.38, "modifiers": ("sunz_corrected",)},
              "refl_1_60um_nom": {"name": "I03", "wavelength": 1.61, "modifiers": ("sunz_corrected",)}
              }
}


def _get_sensor(sensor: str) -> str:
    """Get the sensor."""
    for k, v in SENSORS.items():
        if k in sensor:
            return v
    raise ValueError(f"Unknown sensor '{sensor}'")


def _get_platform(platform: str) -> str:
    """Get the platform."""
    for k, v in PLATFORMS.items():
        if k in platform:
            return v
    return platform


def _get_rows_per_scan(sensor: str) -> Optional[int]:
    """Get number of rows per scan."""
    for k, v in ROWS_PER_SCAN.items():
        if sensor.startswith(k):
            return v
    return None


def _scale_data(data_arr: xr.DataArray | int, scale_factor: float, add_offset: float) -> xr.DataArray:
    """Scale data, if needed."""
    scaling_needed = not (scale_factor == 1.0 and add_offset == 0.0)
    if scaling_needed:
        data_arr = data_arr * np.float32(scale_factor) + np.float32(add_offset)
    return data_arr


class _CLAVRxHelper:
    """A base class for the CLAVRx File Handlers."""

    @staticmethod
    def _get_nadir_resolution(sensor, filename_info_resolution):
        """Get nadir resolution."""
        for k, v in NADIR_RESOLUTION.items():
            if sensor.startswith(k):
                return v
        if filename_info_resolution is None:
            return None
        if isinstance(filename_info_resolution, str) and filename_info_resolution.startswith("m"):
            return int(filename_info_resolution[:-1])
        else:
            return int(filename_info_resolution)

    @staticmethod
    def _remove_attributes(attrs: dict) -> dict:
        """Remove attributes that described data before scaling."""
        old_attrs = ["unscaled_missing", "SCALED_MIN", "SCALED_MAX",
                     "SCALED_MISSING"]

        for attr_key in old_attrs:
            attrs.pop(attr_key, None)
        return attrs

    @staticmethod
    def _get_data(data, dataset_id: dict) -> xr.DataArray:
        """Get a dataset."""
        if dataset_id.get("resolution"):
            data.attrs["resolution"] = dataset_id["resolution"]

        attrs = data.attrs.copy()

        # don't need these attributes after applied.
        factor = attrs.pop("scale_factor", (np.ones(1, dtype=data.dtype))[0])
        offset = attrs.pop("add_offset", (np.zeros(1, dtype=data.dtype))[0])
        flag_values = data.attrs.get("flag_values", [None])
        valid_range = attrs.get("valid_range", [None])

        if isinstance(valid_range, np.ndarray):
            valid_range = valid_range.tolist()
            attrs["valid_range"] = valid_range

        flags = not data.attrs.get("SCALED", 1) and any(flag_values)
        if flags:
            fill = attrs.get("_FillValue", None)
            if isinstance(flag_values, np.ndarray) or isinstance(flag_values, list):
                data = data.where((data >= flag_values[0]) & (data <= flag_values[-1]), fill)
        else:
            fill = attrs.pop("_FillValue", None)
            data = data.where(data != fill)
            data = _scale_data(data, factor, offset)

            if valid_range[0] is not None:
                valid_min = _scale_data(valid_range[0], factor, offset)
                valid_max = _scale_data(valid_range[1], factor, offset)
                data = data.where((data >= valid_min) & (data <= valid_max))
                attrs["valid_range"] = [valid_min, valid_max]

        data.attrs = _CLAVRxHelper._remove_attributes(attrs)

        return data

    @staticmethod
    def _area_extent(x, y, h: float):
        x_l = h * x[0]
        x_r = h * x[-1]
        y_l = h * y[-1]
        y_u = h * y[0]
        ncols = x.shape[0]
        nlines = y.shape[0]
        x_half = (x_r - x_l) / (ncols - 1) / 2.
        y_half = (y_u - y_l) / (nlines - 1) / 2.
        area_extent = (x_l - x_half, y_l - y_half, x_r + x_half, y_u + y_half)
        return area_extent, ncols, nlines

    @staticmethod
    def _read_pug_fixed_grid(projection_coordinates: netCDF4.Variable, distance_multiplier=1.0) -> dict:
        """Read from recent PUG format, where axes are in meters."""
        a = projection_coordinates.semi_major_axis
        h = projection_coordinates.perspective_point_height
        b = projection_coordinates.semi_minor_axis

        lon_0 = projection_coordinates.longitude_of_projection_origin
        sweep_axis = projection_coordinates.sweep_angle_axis[0]

        proj_dict = {"a": float(a) * distance_multiplier,
                     "b": float(b) * distance_multiplier,
                     "lon_0": float(lon_0),
                     "h": float(h) * distance_multiplier,
                     "proj": "geos",
                     "units": "m",
                     "sweep": sweep_axis}
        return proj_dict

    @staticmethod
    def _find_input_nc(filename: str, sensor: str, l1b_base: str) -> str:
        dirname = os.path.dirname(filename)
        l1b_filename = os.path.join(dirname, l1b_base + ".nc")
        if os.path.exists(l1b_filename):
            return str(l1b_filename)

        if sensor == "AHI":
            glob_pat = os.path.join(dirname, l1b_base + "*R20*.nc")
        else:
            glob_pat = os.path.join(dirname, l1b_base + "*.nc")

        LOG.debug("searching for {0}".format(glob_pat))
        found_l1b_filenames = list(glob(glob_pat))
        if len(found_l1b_filenames) == 0:
            fp = os.path.join(dirname, l1b_base)
            raise IOError(f"Missing navigation donor {fp}")
        LOG.debug("Candidate nav donors: {0}".format(repr(found_l1b_filenames)))
        return found_l1b_filenames[0]

    @staticmethod
    def _read_axi_fixed_grid(filename: str, sensor: str, l1b_attr) -> geometry.AreaDefinition:
        """Read a fixed grid.

        CLAVR-x does not transcribe fixed grid parameters to its output
        We have to recover that information from the original input file,
        which is partially named as L1B attribute

        example attributes found in L2 CLAVR-x files:
        sensor = "AHI" ;
        platform = "HIM8" ;
        FILENAME = "clavrx_H08_20180719_1300.level2.hdf" ;
        L1B = "clavrx_H08_20180719_1300" ;
        """
        LOG.debug(f"looking for corresponding input file for {l1b_attr}"
                  " to act as fixed grid navigation donor")
        l1b_path = _CLAVRxHelper._find_input_nc(filename, sensor, l1b_attr)
        LOG.info(f"CLAVR-x does not include fixed-grid parameters, use input file {l1b_path} as donor")
        l1b = netCDF4.Dataset(l1b_path)
        proj = None
        proj_var = l1b.variables.get("Projection", None)
        if proj_var is not None:
            # hsd2nc input typically used by CLAVR-x uses old-form km for axes/height
            LOG.debug("found hsd2nc-style draft PUG fixed grid specification")
            proj = _CLAVRxHelper._read_pug_fixed_grid(proj_var, 1000.0)
        if proj is None:  # most likely to come into play for ABI cases
            proj_var = l1b.variables.get("goes_imager_projection", None)
            if proj_var is not None:
                LOG.debug("found cmip-style final PUG fixed grid specification")
                proj = _CLAVRxHelper._read_pug_fixed_grid(proj_var)
        if not proj:
            raise ValueError(f"Unable to recover projection information for {filename}")

        h = float(proj["h"])
        x, y = l1b["x"], l1b["y"]
        area_extent, ncols, nlines = _CLAVRxHelper._area_extent(x, y, h)

        area = geometry.AreaDefinition(
          f"{sensor}_geos",
          f"{sensor.upper()} L2 file area",
          f"{sensor}_geos",
          proj,
          ncols,
          nlines,
          area_extent)

        return area

    @staticmethod
    def get_metadata(sensor: str, platform: str, attrs: dict, ds_info: dict) -> dict:
        """Get metadata."""
        attr_info = {}
        attr_info.update(attrs)
        attr_info.update(ds_info)

        flag_meanings = attr_info.get("flag_meanings", None)
        if not attr_info.get("SCALED", 1) and not flag_meanings:
            attr_info["flag_meanings"] = "<flag_meanings_unknown>"
            attr_info.setdefault("flag_values", [None])
        elif not attr_info.get("SCALED", 1) and isinstance(flag_meanings, str):
            attr_info["flag_meanings"] = flag_meanings.split("  ")
        u = attr_info.get("units")
        if u in CF_UNITS:
            # CF compliance
            attr_info["units"] = CF_UNITS[u]
            if u.lower() == "none":
                attr_info["units"] = "1"
        attr_info["sensor"] = sensor
        attr_info["platform_name"] = platform
        rps = _get_rows_per_scan(sensor)
        if rps:
            attr_info["rows_per_scan"] = rps
        attr_info["reader"] = "clavrx"

        return attr_info


class CLAVRXHDF4FileHandler(HDF4FileHandler, _CLAVRxHelper):
    """A file handler for CLAVRx files."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init method."""
        super(CLAVRXHDF4FileHandler, self).__init__(filename,
                                                    filename_info,
                                                    filetype_info)

        self.sensor = _get_sensor(self.file_content.get("/attr/sensor"))
        self.platform = _get_platform(self.file_content.get("/attr/platform"))
        self.resolution = _CLAVRxHelper._get_nadir_resolution(self.sensor,
                                                              self.filename_info.get("resolution"))

    @property
    def start_time(self):
        """Get the start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get the end time."""
        return self.filename_info.get("end_time", self.start_time)

    def get_dataset(self, dataset_id, ds_info):
        """Get a dataset for Polar Sensors."""
        var_name = ds_info.get("file_key", dataset_id["name"])
        data = self[var_name]
        data = _CLAVRxHelper._get_data(data, dataset_id)
        data.attrs = _CLAVRxHelper.get_metadata(self.sensor, self.platform,
                                                data.attrs, ds_info)
        return data

    def _available_aliases(self, ds_info, current_var):
        """Add alias if there is a match."""
        new_info = ds_info.copy()
        alias_info = CHANNEL_ALIASES.get(self.sensor).get(current_var, None)
        if alias_info is not None:
            alias_info.update({"file_key": current_var})
            new_info.update(alias_info)
            yield True, new_info

    def available_datasets(self, configured_datasets=None):
        """Add more information if this reader can provide it."""
        handled_variables = set()
        for is_avail, ds_info in (configured_datasets or []):
            # some other file handler knows how to load this
            if is_avail is not None:
                yield is_avail, ds_info

            new_info = ds_info.copy()  # don't change input
            this_res = ds_info.get("resolution")
            var_name = ds_info.get("file_key", ds_info["name"])
            matches = self.file_type_matches(ds_info["file_type"])
            # we can confidently say that we can provide this dataset and can
            # provide more info
            if matches and var_name in self and this_res != self.resolution:
                handled_variables.add(var_name)
                new_info["resolution"] = self.resolution
                if self._is_polar():
                    new_info["coordinates"] = ds_info.get("coordinates", ("longitude", "latitude"))
                yield True, new_info
            elif is_avail is None:
                # if we didn't know how to handle this dataset and no one else did
                # then we should keep it going down the chain
                yield is_avail, ds_info

        # get data from file dynamically
        yield from self._dynamic_datasets()

    def _dynamic_datasets(self):
        """Get data from file and build aliases."""
        for var_name, val in self.file_content.items():
            if isinstance(val, SDS):
                ds_info = {
                    "file_type": self.filetype_info["file_type"],
                    "resolution": self.resolution,
                    "name": var_name,
                }
                if self._is_polar():
                    ds_info["coordinates"] = ["longitude", "latitude"]

                # always yield what we have
                yield True, ds_info
                if CHANNEL_ALIASES.get(self.sensor) is not None:
                    # yield variable as it is
                    # yield any associated aliases
                    yield from self._available_aliases(ds_info, var_name)

    def get_shape(self, dataset_id, ds_info):
        """Get the shape."""
        var_name = ds_info.get("file_key", dataset_id["name"])
        return self[var_name + "/shape"]

    def _is_polar(self):
        l1b_att, inst_att = (str(self.file_content.get("/attr/L1B", None)),
                             str(self.file_content.get("/attr/sensor", None)))

        return (inst_att != "AHI" and "GOES" not in inst_att) or (l1b_att is None)

    def get_area_def(self, key):
        """Get the area definition of the data at hand."""
        if self._is_polar():  # then it doesn't have a fixed grid
            return super(CLAVRXHDF4FileHandler, self).get_area_def(key)

        l1b_att = str(self.file_content.get("/attr/L1B", None))
        area_def = _CLAVRxHelper._read_axi_fixed_grid(self.filename, self.sensor, l1b_att)
        return area_def


class CLAVRXNetCDFFileHandler(_CLAVRxHelper, BaseFileHandler):
    """File Handler for CLAVRX netcdf files."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init method."""
        super(CLAVRXNetCDFFileHandler, self).__init__(filename,
                                                      filename_info,
                                                      filetype_info,
                                                      )

        self.nc = xr.open_dataset(filename,
                                  decode_cf=True,
                                  mask_and_scale=False,
                                  decode_coords=True,
                                  chunks=CHUNK_SIZE)
        # y,x is used in satpy, bands rather than channel using in xrimage
        self.nc = self.nc.rename_dims({"scan_lines_along_track_direction": "y",
                                       "pixel_elements_along_scan_direction": "x"})

        self.platform = _get_platform(
            self.filename_info.get("platform_shortname", None))
        self.sensor = _get_sensor(self.nc.attrs.get("sensor", None))
        self.resolution = _CLAVRxHelper._get_nadir_resolution(self.sensor,
                                                              self.filename_info.get("resolution"))
        # coordinates need scaling and valid_range (mask_and_scale won't work on valid_range)
        self.nc.coords["latitude"] = _CLAVRxHelper._get_data(self.nc.coords["latitude"],
                                                             {"name": "latitude"})
        self.nc.coords["longitude"] = _CLAVRxHelper._get_data(self.nc.coords["longitude"],
                                                              {"name": "longitude"})

    def _dynamic_dataset_info(self, var_name):
        """Set data name and, if applicable, aliases."""
        ds_info = {
            "file_type": self.filetype_info["file_type"],
            "name": var_name,
        }
        yield True, ds_info

        if CHANNEL_ALIASES.get(self.sensor) is not None:
            alias_info = ds_info.copy()
            channel_info = CHANNEL_ALIASES.get(self.sensor).get(var_name, None)
            if channel_info is not None:
                channel_info["file_key"] = var_name
                alias_info.update(channel_info)
                yield True, alias_info

    @staticmethod
    def _is_2d_yx_data_array(data_arr):
        has_y_dim = data_arr.dims[0] == "y"
        has_x_dim = data_arr.dims[1] == "x"
        return has_y_dim and has_x_dim

    def _available_file_datasets(self, handled_vars):
        """Metadata for available variables other than BT."""
        possible_vars = list(self.nc.items()) + list(self.nc.coords.items())
        for var_name, data_arr in possible_vars:
            if var_name in handled_vars:
                continue
            if data_arr.ndim != 2:
                # we don't currently handle non-2D variables
                continue
            if not self._is_2d_yx_data_array(data_arr):
                # we need 'traditional' y/x dimensions currently
                continue

            yield from self._dynamic_dataset_info(var_name)

    def available_datasets(self, configured_datasets=None):
        """Dynamically discover what variables can be loaded from this file.

        See :meth:`satpy.readers.core.file_handlers.BaseFileHandler.available_datasets`
        for more information.

        """
        handled_vars = set()
        for is_avail, ds_info in (configured_datasets or []):
            if is_avail is not None:
                # some other file handler said it has this dataset
                # we don't know any more information than the previous
                # file handler so let's yield early
                yield is_avail, ds_info

            matches = self.file_type_matches(ds_info["file_type"])
            if matches and ds_info.get("resolution") != self.resolution:
                # reader knows something about this dataset (file type matches)
                # add any information that this reader can add.
                new_info = ds_info.copy()
                if self.resolution is not None:
                    new_info["resolution"] = self.resolution
                handled_vars.add(ds_info["name"])
                yield True, new_info
        yield from self._available_file_datasets(handled_vars)

    def _is_polar(self):
        l1b_att, inst_att = (str(self.nc.attrs.get("L1B", None)),
                             str(self.nc.attrs.get("sensor", None)))

        return (inst_att not in ["ABI", "AHI"] and "GOES" not in inst_att) or (l1b_att is None)

    def get_area_def(self, key):
        """Get the area definition of the data at hand."""
        if self._is_polar():  # then it doesn't have a fixed grid
            return super(CLAVRXNetCDFFileHandler, self).get_area_def(key)

        l1b_att = str(self.nc.attrs.get("L1B", None))
        return _CLAVRxHelper._read_axi_fixed_grid(self.filename, self.sensor, l1b_att)

    def get_dataset(self, dataset_id, ds_info):
        """Get a dataset for supported geostationary sensors."""
        var_name = ds_info.get("file_key", dataset_id["name"])
        data = self[var_name]
        data = _CLAVRxHelper._get_data(data, dataset_id)
        data.attrs = _CLAVRxHelper.get_metadata(self.sensor, self.platform,
                                                data.attrs, ds_info)
        return data

    def __getitem__(self, item):
        """Wrap around `self.nc[item]`."""
        # Check if "item" is an alias:
        data = self.nc[item]
        return data



"""Testing helpers for satpy."""

from contextlib import contextmanager
from unittest import mock

import pytest

import satpy.scene


@contextmanager
def fake_satpy_reading(scene_dict):
    """Fake the satpy reading and populate the returned scene with the contents of *scene_dict*.

    This allows users to test their programs that use satpy without actually needing to read files, eg::

        scene_dict = {channel: somedata}

        with fake_satpy_reading(scene_dict):
            scene = Scene(input_files, reader="dummy_reader")
            scene.load([channel])

    """
    with pytest.MonkeyPatch().context() as monkeypatch:
        reader_instance = mock.Mock()
        reader_instance.sensor_names = ["dummy_sensor"]
        fake_load_readers = mock.Mock()
        fake_load_readers.return_value = {"dummy_reader": reader_instance}
        monkeypatch.setattr(satpy.scene, "load_readers", fake_load_readers)

        def fake_load(self, channels):
            for channel in channels:
                self[channel] = scene_dict[channel]


        monkeypatch.setattr(satpy.scene.Scene, "load", fake_load)
        yield



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""GeoTIFF writer objects for creating GeoTIFF files from :class:`xarray.DataArray` objects."""
from __future__ import annotations

import logging
import typing

import dask.array as da
import numpy as np

# make sure we have rasterio even though we don't use it until trollimage
# saves the image
import rasterio  # noqa
from numpy.typing import DTypeLike
from trollimage.colormap import Colormap
from trollimage.xrimage import XRImage

from satpy.writers.core.image import ImageWriter

if typing.TYPE_CHECKING:
    from os import PathLike
    from typing import Any

    from dask.delayed import Delayed

LOG = logging.getLogger(__name__)


class GeoTIFFWriter(ImageWriter):
    """Writer to save GeoTIFF images.

    Basic example from Scene:

        >>> scn.save_datasets(writer='geotiff')

    By default the writer will use the :class:`~satpy.enhancements.enhancer.Enhancer` class to
    linear stretch the data (see :doc:`../enhancements`).
    To get Un-enhanced images ``enhance=False`` can be specified which will
    write a geotiff with the data type of the dataset. The fill value defaults
    to the the datasets ``"_FillValue"`` attribute if not ``None`` and no value is
    passed to ``fill_value`` for integer data. In case of float data if ``fill_value``
    is not passed NaN will be used. If a geotiff with a
    certain datatype is desired for example 32 bit floating point geotiffs:

        >>> scn.save_datasets(writer='geotiff', dtype=np.float32, enhance=False)

    To add custom metadata use `tags`:

        >>> scn.save_dataset(dataset_name, writer='geotiff',
        ...                  tags={'offset': 291.8, 'scale': -0.35})

    Images are tiled by default. To create striped TIFF files ``tiled=False`` can be specified:

        >>> scn.save_datasets(writer='geotiff', tiled=False)

    For performance tips on creating geotiffs quickly and making them smaller
    see the :ref:`faq`.

    """

    GDAL_OPTIONS = ("tfw",
                    "rpb",
                    "rpctxt",
                    "interleave",
                    "tiled",
                    "blockxsize",
                    "blockysize",
                    "nbits",
                    "compress",
                    "num_threads",
                    "predictor",
                    "discard_lsb",
                    "sparse_ok",
                    "jpeg_quality",
                    "jpegtablesmode",
                    "zlevel",
                    "photometric",
                    "alpha",
                    "profile",
                    "bigtiff",
                    "pixeltype",
                    "copy_src_overviews",
                    # COG driver options (different from GTiff above)
                    "blocksize",
                    "resampling",
                    "quality",
                    "level",
                    "overview_resampling",
                    "warp_resampling",
                    "overview_compress",
                    "overview_quality",
                    "overview_predictor",
                    "tiling_scheme",
                    "zoom_level_strategy",
                    "target_srs",
                    "res",
                    "extent",
                    "aligned_levels",
                    "add_alpha",
                    )

    def __init__(self, dtype=None, tags=None, **kwargs):
        """Init the writer."""
        super().__init__(default_config_filename="writers/geotiff.yaml", **kwargs)
        self.dtype = self.info.get("dtype") if dtype is None else dtype
        self.tags = self.info.get("tags", None) if tags is None else tags
        if self.tags is None:
            self.tags = {}
        elif not isinstance(self.tags, dict):
            # if it's coming from a config file
            self.tags = dict(tuple(x.split("=")) for x in self.tags.split(","))

        # GDAL specific settings
        self.gdal_options = {}
        for k in self.GDAL_OPTIONS:
            if k in kwargs or k in self.info:
                self.gdal_options[k] = kwargs.get(k, self.info[k])

    @classmethod
    def separate_init_kwargs(cls, kwargs):
        """Separate the init keyword args."""
        # FUTURE: Don't pass Scene.save_datasets kwargs to init and here
        init_kwargs, kwargs = super(GeoTIFFWriter, cls).separate_init_kwargs(
            kwargs)
        for kw in ["dtype", "tags"]:
            if kw in kwargs:
                init_kwargs[kw] = kwargs.pop(kw)

        return init_kwargs, kwargs

    def save_image(  # noqa: D417
            self,
            img: XRImage,
            filename: str | None = None,
            compute: bool = True,
            dtype: DTypeLike | None = None,
            fill_value: int | float | None = None,
            keep_palette: bool = False,
            cmap: Colormap | None = None,
            tags: dict[str, Any] | None = None,
            overviews: list[int] | None = None,
            overviews_minsize: int = 256,
            overviews_resampling: str | None = None,
            include_scale_offset: bool = False,
            scale_offset_tags: tuple[str, str] | None = None,
            colormap_tag: str | None = None,
            driver: str | None = None,
            tiled: bool = True,
            **kwargs
    ) -> list[da.Array | Delayed] | tuple[list[da.Array], list[Any]] | list[str | PathLike | None]:
        """Save the image to the given ``filename`` in geotiff_ format.

        Note this writer requires the ``rasterio`` library to be installed.

        Args:
            img: Data to save to geotiff.
            filename: Filename to save the image to. Defaults to
                ``filename`` passed during writer creation. Unlike the
                creation ``filename`` keyword argument, this filename does not
                get formatted with data attributes.
            compute: Compute dask arrays and save the image
                immediately. If ``False`` then the return value can be passed
                to :func:`~satpy.writers.core.compute.compute_writer_results` to do the
                computation. This is useful when multiple images may share
                input calculations where dask can benefit from not repeating
                them multiple times. Defaults to ``True`` in the writer by
                itself, but is typically passed as ``False`` by callers where
                calculations can be combined.
            dtype: Numpy data type to save the image as.
                Defaults to 8-bit unsigned integer (``np.uint8``) or the data
                type of the data to be saved if ``enhance=False``. If the
                ``dtype`` argument is provided during writer creation then
                that will be used as the default.
            fill_value: Value to use where data values are
                NaN/null. If this is specified in the writer configuration
                file that value will be used as the default.
            keep_palette: Save palette/color table to geotiff.
                To be used with images that were palettized with the
                "palettize" enhancement. Setting this to ``True`` will cause
                the colormap of the image to be written as a "color table" in
                the output geotiff and the image data values will represent
                the index values in to that color table. By default, this will
                use the colormap used in the "palettize" operation.
                See the ``cmap`` option for other options. This option defaults
                to ``False`` and palettized images will be converted to RGB/A.
            cmap: Colormap to save
                as a color table in the output geotiff. See ``keep_palette``
                for more information. Defaults to the palette of the provided
                ``img`` object. The colormap's range should be set to match
                the index range of the palette
                (ex. `cmap.set_range(0, len(colors))`).
            tags: Extra metadata to store in geotiff.
            overviews: The reduction factors of the overviews to include
                in the image, eg::

                    scn.save_datasets(overviews=[2, 4, 8, 16])

                If provided as an empty list, then levels will be
                computed as powers of two until the last level has less
                pixels than `overviews_minsize`.
                Default is to not add overviews.
            overviews_minsize: Minimum number of pixels for the smallest
                overview size generated when `overviews` is auto-generated.
                Defaults to 256.
            overviews_resampling: Resampling method
                to use when generating overviews. This must be the name of an
                enum value from :class:`rasterio.enums.Resampling` and
                only takes effect if the `overviews` keyword argument is
                provided. Common values include `nearest` (default),
                `bilinear`, `average`, and many others. See the rasterio
                documentation for more information.
            scale_offset_tags: If set, include inclusion of
                scale and offset in the GeoTIFF headers in the GDALMetaData
                tag.  The value of this argument should be a keyword argument
                ``(scale_label, offset_label)``, for example, ``("scale",
                "offset")``, indicating the labels to be used.
            colormap_tag: If set and the image being saved was
                colorized or palettized then a comma-separated version of the
                colormap is saved to a custom geotiff tag with the provided
                name. See :meth:`trollimage.colormap.Colormap.to_csv` for more
                information.
            driver: Name of GDAL driver to use to save the
                geotiff. If not specified or None (default) the "GTiff" driver
                is used. Another common option is "COG" for Cloud Optimized
                GeoTIFF. See GDAL documentation for more information.
            tiled: For performance this defaults to ``True``.
                Pass ``False`` to created striped TIFF files.
            include_scale_offset: Deprecated.
                Use ``scale_offset_tags=("scale", "offset")`` to include scale
                and offset tags.

        .. _geotiff: http://trac.osgeo.org/geotiff/

        """
        filename = filename or self.get_filename(**img.data.attrs)

        gdal_options = self._get_gdal_options(kwargs)
        if fill_value is None:
            # fall back to fill_value from configuration file
            fill_value = self.info.get("fill_value")

        dtype = dtype if dtype is not None else self.dtype
        if dtype is None and self.enhancer is not False:
            dtype = np.uint8
        elif dtype is None:
            dtype = img.data.dtype.type

        if "alpha" in kwargs:
            raise ValueError(
                "Keyword 'alpha' is automatically set based on 'fill_value' "
                "and should not be specified")
        if np.issubdtype(dtype, np.floating):
            if img.mode != "L":
                raise ValueError("Image must be in 'L' mode for floating "
                                 "point geotiff saving")
            if fill_value is None:
                LOG.debug("Alpha band not supported for float geotiffs, "
                          "setting fill value to 'NaN'")
                fill_value = np.nan
        if keep_palette and cmap is None and img.palette is not None:
            from satpy.enhancements.colormap import create_colormap
            cmap = create_colormap({"colors": img.palette})
            cmap.set_range(0, len(img.palette) - 1)

        if tags is None:
            tags = {}
        tags.update(self.tags)

        res = img.save(filename, fformat="tif", driver=driver,
                       fill_value=fill_value,
                       dtype=dtype, compute=compute,
                       keep_palette=keep_palette, cmap=cmap,
                       tags=tags, include_scale_offset_tags=include_scale_offset,
                       scale_offset_tags=scale_offset_tags,
                       colormap_tag=colormap_tag,
                       overviews=overviews,
                       overviews_resampling=overviews_resampling,
                       overviews_minsize=overviews_minsize,
                       tiled=tiled,
                       **gdal_options)
        return _trollimage_res_to_satpy_res(res)

    def _get_gdal_options(self, kwargs):
        # Update global GDAL options with these specific ones
        gdal_options = self.gdal_options.copy()
        for k in kwargs:
            if k in self.GDAL_OPTIONS:
                gdal_options[k] = kwargs[k]
        return gdal_options


def _trollimage_res_to_satpy_res(res):
    # in old trollimage <1.27.0: result is either None or (sources_list, targets_list) or (source, target)
    # in new trollimage: result is either str or Path or (sources_list, targets_list)
    if isinstance(res, tuple):
        if isinstance(res[0], da.Array):
            # convert old trollimage single (source, target) to ([source], [target])
            res = ([res[0]], [res[1]])
        return res
    if isinstance(res, list) and len(res) == 2 and isinstance(res[0], tuple):
        # old trollimage with dask-based tags specified (see ninjogeotiff writer)
        # returns [(tuple of sources), (tuple of targets)], this swaps those types:
        return list(res[0]), list(res[1])
    return [res]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Modifier classes for corrections based on sun and other angles."""

from __future__ import annotations

import logging

import numpy as np

from satpy.modifiers import ModifierBase
from satpy.modifiers.angles import sunzen_corr_cos, sunzen_reduction
from satpy.utils import atmospheric_path_length_correction

logger = logging.getLogger(__name__)


class SunZenithCorrectorBase(ModifierBase):
    """Base class for sun zenith correction modifiers."""

    def __init__(self, max_sza=95.0, **kwargs):  # noqa: D417
        """Collect custom configuration values.

        Args:
            max_sza (float): Maximum solar zenith angle in degrees that is
                considered valid and correctable. Default 95.0.

        """
        self.max_sza = max_sza
        self.max_sza_cos = np.cos(np.deg2rad(max_sza)) if max_sza is not None else None
        super(SunZenithCorrectorBase, self).__init__(**kwargs)

    def __call__(self, projectables, **info):
        """Generate the composite."""
        projectables = self.match_data_arrays(list(projectables) + list(info.get("optional_datasets", [])))
        vis = projectables[0]
        if vis.attrs.get("sunz_corrected"):
            logger.debug("Sun zenith correction already applied")
            return vis

        logger.debug("Applying sun zen correction")
        if not info.get("optional_datasets"):
            # we were not given SZA, generate cos(SZA)
            logger.debug("Computing sun zenith angles.")
            from .angles import get_cos_sza
            coszen = get_cos_sza(vis)
            if self.max_sza is not None:
                coszen = coszen.where(coszen >= self.max_sza_cos)
        else:
            # we were given the SZA, calculate the cos(SZA)
            coszen = np.cos(np.deg2rad(projectables[1]))

        proj = self._apply_correction(vis, coszen)
        proj.attrs = vis.attrs.copy()
        self.apply_modifier_info(vis, proj)
        return proj

    def _apply_correction(self, proj, coszen):
        raise NotImplementedError("Correction method shall be defined!")


class SunZenithCorrector(SunZenithCorrectorBase):
    """Standard sun zenith correction using ``1 / cos(sunz)``.

    In addition to adjusting the provided reflectances by the cosine of the
    solar zenith angle, this modifier forces all reflectances beyond a
    solar zenith angle of ``max_sza`` to 0. It also gradually reduces the
    amount of correction done between ``correction_limit`` and ``max_sza``. If
    ``max_sza`` is ``None`` then a constant correction is applied to zenith
    angles beyond ``correction_limit``.

    To set ``max_sza`` to ``None`` in a YAML configuration file use:

    .. code-block:: yaml

      sunz_corrected:
        modifier: !!python/name:satpy.modifiers.SunZenithCorrector
        max_sza: !!null
        optional_prerequisites:
        - solar_zenith_angle

    """

    def __init__(self, correction_limit=88., **kwargs):  # noqa: D417
        """Collect custom configuration values.

        Args:
            correction_limit (float): Maximum solar zenith angle to apply the
                correction in degrees. If ``max_sza`` is ``None``, pixels beyond this limit have a
                constant correction applied. Otherwise, the correction is gradually reduced to 0 at
                ``max_sza``. Default 88.
            max_sza (float): Maximum solar zenith angle in degrees that is
                considered valid and correctable. Default 95.0.

        """
        self.correction_limit = correction_limit
        super(SunZenithCorrector, self).__init__(**kwargs)

    def _apply_correction(self, proj, coszen):
        logger.debug("Apply the standard sun-zenith correction [1/cos(sunz)]")
        res = proj.copy()
        res.data = sunzen_corr_cos(proj.data, coszen.data, limit=self.correction_limit, max_sza=self.max_sza)
        return res


class EffectiveSolarPathLengthCorrector(SunZenithCorrectorBase):
    """Special sun zenith correction with the method proposed by Li and Shibata.

    (2006): https://doi.org/10.1175/JAS3682.1

    In addition to adjusting the provided reflectances by the cosine of the
    solar zenith angle, this modifier forces all reflectances beyond a
    solar zenith angle of `max_sza` to 0 to reduce noise in the final data.
    It also gradually reduces the amount of correction done between
    ``correction_limit`` and ``max_sza``. If ``max_sza`` is ``None`` then a
    constant correction is applied to zenith angles beyond
    ``correction_limit``.

    To set ``max_sza`` to ``None`` in a YAML configuration file use:

    .. code-block:: yaml

      effective_solar_pathlength_corrected:
        modifier: !!python/name:satpy.modifiers.EffectiveSolarPathLengthCorrector
        max_sza: !!null
        optional_prerequisites:
        - solar_zenith_angle

    """

    def __init__(self, correction_limit=88., **kwargs):  # noqa: D417
        """Collect custom configuration values.

        Args:
            correction_limit (float): Maximum solar zenith angle to apply the
                correction in degrees. If ``max_sza`` is ``None``, pixels beyond this limit have a
                constant correction applied. Otherwise, the correction is gradually reduced to 0 at
                ``max_sza``. Default 88.
            max_sza (float): Maximum solar zenith angle in degrees that is
                considered valid and correctable. Default 95.0.

        """
        self.correction_limit = correction_limit
        super(EffectiveSolarPathLengthCorrector, self).__init__(**kwargs)

    def _apply_correction(self, proj, coszen):
        logger.debug("Apply the effective solar atmospheric path length correction method by Li and Shibata")
        return atmospheric_path_length_correction(proj, coszen, limit=self.correction_limit, max_sza=self.max_sza)


class SunZenithReducer(SunZenithCorrectorBase):
    """Reduce signal strength at large sun zenith angles.

    Within a given sunz interval [correction_limit, max_sza] the strength of the signal is reduced following the
    formula:

      res = signal * reduction_factor

    where reduction_factor is a pixel-level value ranging from 0 to 1 within the sunz interval.

    The `strength` parameter can be used for a non-linear reduction within the sunz interval. A strength larger
    than 1.0 will decelerate the signal reduction towards the sunz interval extremes, whereas a strength
    smaller than 1.0 will accelerate the signal reduction towards the sunz interval extremes.

    """

    def __init__(self, correction_limit=80., max_sza=90, strength=1.3, **kwargs):  # noqa: D417
        """Collect custom configuration values.

        Args:
            correction_limit (float): Solar zenith angle in degrees where to start the signal reduction.
            max_sza (float): Maximum solar zenith angle in degrees where to apply the signal reduction. Beyond
                             this solar zenith angle the signal will become zero.
            strength (float): The strength of the non-linear signal reduction.

        """
        self.correction_limit = correction_limit
        self.strength = strength
        super(SunZenithReducer, self).__init__(max_sza=max_sza, **kwargs)
        if self.max_sza is None:
            raise ValueError("`max_sza` must be defined when using the SunZenithReducer.")

    def _apply_correction(self, proj, coszen):
        logger.debug(f"Applying sun-zenith signal reduction with correction_limit {self.correction_limit} deg,"
                     f" strength {self.strength}, and max_sza {self.max_sza} deg.")
        res = proj.copy()
        sunz = np.rad2deg(np.arccos(coszen.data))
        res.data = sunzen_reduction(proj.data, sunz,
                                    limit=self.correction_limit,
                                    max_sza=self.max_sza,
                                    strength=self.strength)
        return res
